# -*- coding: utf-8 -*-
from contextlib import contextmanager
# from snipy.odict import odict

_stack = [None]


class Scope(object):
    def __init__(self, name):
        self.name = name


@contextmanager
def vscope(name):

    # get current scope or new scope
    # append hierachical information
    sc = Scope(name)
    _stack.append(sc)

    try:
        yield sc
    finally:
        _stack.pop(-1)


def get_scope():
    return _stack[-1]




# -*- coding: utf-8 -*-
import os
from codecs import open
import scandir
from snipy.ilogging import logg
import numpy as np


def mkdir_if_not(filepath, ispath=False):
    """
    path 부분이 없으면 mkdir 을 한다.
    :param filepath: 파일 패쓰
    :return: filpath 그대로 리턴
    """
    if not ispath:
        p, _ = os.path.split(filepath)
    else:
        p = filepath

    if not p:
        return filepath
    if not os.path.exists(p):
        # M.info('%s not exist, trying mkdir ', p)
        try:
            os.makedirs(p)
        except FileExistsError as e:
            logg.warn(str(e))

    return filepath


def readlines(filepath):
    """
    read lines from a textfile
    :param filepath:
    :return: list[line]
    """
    with open(filepath, 'rt') as f:
        lines = f.readlines()
        lines = map(str.strip, lines)
        lines = [l for l in lines if l]
    return lines


def writelines(filepath, lines):
    mkdir_if_not(filepath)
    with open(filepath, 'wt') as f:
        for l in lines:
            f.write(l + '\n')


def readtxt(filepath):
    """ read file as is"""
    with open(filepath, 'rt') as f:
        lines = f.readlines()
    return ''.join(lines)


def writetxt(filepath, txt):
    mkdir_if_not(filepath)
    with open(filepath, 'wt') as f:
        f.write(txt)


def savefile(obj, filepath, compress=True):
    """
    파일 있으면 덮어씀
    :param obj:
    :param str filepath:
    :param compress:
    :return:
    """
    try:
        import cPickle as pickle
    except Exception:
        import pickle
    import joblib

    # 일단 임시 파일에 저장.
    tmpfile = filepath + '.tmp'
    mkdir_if_not(tmpfile)
    if compress:
        joblib.dump(obj, tmpfile, compress=3, cache_size=100, protocol=pickle.HIGHEST_PROTOCOL)
    else:
        joblib.dump(obj, tmpfile, compress=0)

    os.rename(tmpfile, filepath)

    return obj


def loadfile(filepath, mmap_mode=None):
    """
    :param filepath:
    :param mmap_mode: {None, ‘r+’, ‘r’, ‘w+’, ‘c’} see. joblib.load
    :return:
    """
    import joblib

    try:
        return joblib.load(filepath, mmap_mode=mmap_mode)
    except IOError:
        return None


def latest_file(fpattern, matchfun=None):
    """
    regular format으로 조회한 파일리스트중에 가장최신 파일 path리턴 file modified time 조회 비교
    :param function matchfun:
    :param fpattern: 파일 패턴 ( ex: data/some*.txt)
    :return: data/somelatest.txt, None 이면 파일이 없는 것
    """
    import glob
    # matchfun = matchfun or glob.glob
    files = glob.glob(fpattern)
    if matchfun:
        files = filter(matchfun, files)

    latest, maxtime = None, 0
    for f in files:
        t = os.path.getmtime(f)
        if t > maxtime:
            latest, maxtime = f, t

    return latest


def load_latest(fpattern):
    latest = latest_file(fpattern)
    if latest is None:
        return None
    else:
        return loadfile(latest)


def load_or_run(filepath, fun, *args, **kwargs):
    """
    계산된 결과 파일이 있으면 로딩하고, 없으면 계산후 저장
    ex)
    res = load_or_run('file_loadorsave', funlongtime, ...., force=False)
    :param filepath:
    :param fun:
    :param force:
    :return:
    """
    force = kwargs.pop('force', False)
    compress = kwargs.pop('compress', True)

    if not filepath.startswith('/') or not filepath.startswith('~'):
        filepath = os.path.join('/tmp/snipy/load_or_run/', filepath)

    if not force and os.path.exists(filepath):
        # 저장되어 있는 것 로딩
        mmap_mode = 'r+' if not compress else None
        return loadfile(filepath, mmap_mode=mmap_mode)

    res = fun(*args, **kwargs)
    savefile(res, filepath, compress=compress)

    return res


def readhdf5(f):
    import warnings
    warnings.simplefilter(action='ignore', category=FutureWarning)
    import h5py
    warnings.resetwarnings()

    return h5py.File(f, 'r')


def getpath(f, isfile=True):
    if isfile:
        return os.path.dirname(os.path.realpath(f))
    else:
        return os.path.realpath(f)


# list directory, file, recursive or not, ...
# import scandir

def any_match(fname, patterns, matchfun=None):
    """
    ANY matches?
    :param str fname: file name
    :param list[str] patterns: list of filename pattern. see fnmatch.fnamtch
    :rtype: bool
    """
    return any(fnmatches(fname, patterns, matchfun))


def fnmatches(fname, patterns, matchfun):
    """"
    matches?
    :param fname: file name
    :type fname: str
    :param patterns: list of filename pattern. see fnmatch.fnamtch
    :type patterns: [str]
    :rtype: generator of bool
    """
    import fnmatch
    matchfun = matchfun or fnmatch.fnmatch
    for p in patterns:
        yield matchfun(fname, p)


def listdir(p, match='*', exclude='', listtype='file', matchfun=None):
    """
    list file(or folder) for this path (NOT recursive)
    :param p:
    :param match:
    :param exclude:
    :param listtype: ('file' | 'filepath' |'dir' | 'all')
    :param matchfun: match fun (default fnmatch.fnmatch) True/False = matchfun(name, pattern)
    :rtype:
    """
    if listtype == 'file':
        gen = listfile(p)
    elif listtype == 'filepath':
        gen = listfilepath(p)
    elif listtype == 'dir':
        gen = listfolder(p)
    elif listtype == 'dirpath':
        gen = listfolderpath(p)
    else:  # list file or folder
        gen = (entry.name for entry in scandir.scandir(p))

    return filter_pattern(gen, match, exclude, matchfun)


def filter_pattern(gen, include='*', exclude='', matchfun=None):
    pred = _pred_pattern(include, exclude, matchfun)
    for f in gen:
        if pred(f):
            yield f


def listfile(p):
    """
    generator of list files in the path.
    filenames only
    """
    try:
        for entry in scandir.scandir(p):
            if entry.is_file():
                yield entry.name
    except OSError:
        return


def listfilepath(p):
    """
    generator of list files in the path.
    filenames only
    """
    for entry in scandir.scandir(p):
        if entry.is_file():
            yield entry.path


def listfolder(p):
    """
    generator of list folder in the path.
    folders only
    """
    for entry in scandir.scandir(p):
        if entry.is_dir():
            yield entry.name


def listfolderpath(p):
    """
    generator of list folder in the path.
    folders only
    """
    for entry in scandir.scandir(p):
        if entry.is_dir():
            yield entry.path


def get_match_fun(patterns, patterntype):
    import re

    def _fnmatch(fname):
        return any_match(fname, patterns, patterntype)

    if patterntype != 're':  # usually just fnmatch
        return _fnmatch

    patterns = [re.compile(p) for p in patterns]

    def _re_match(fname):
        for p in patterns:
            if p.match(fname):
                return True
        return False

    return _re_match


def _is_str(x):
    
    try:
        return isinstance(x, (str, basestring))
    except NameError:
        return isinstance(x, str)


def _pred_pattern(match='*', exclude='', patterntype='fnmatch'):

    """ internal use """
    m, x = match, exclude
    if m == '*':
        if not x:
            pred = lambda n: True
        else:
            x = [x] if _is_str(x) else x
            matcher = get_match_fun(x, patterntype)
            pred = lambda n: not matcher(n)
    else:
        m = [m] if _is_str(m) else m
        if not x:
            matcher = get_match_fun(m, patterntype)
            pred = lambda n: matcher(n)
        else:
            x = [x] if _is_str(x) else x
            matcher_m = get_match_fun(m, patterntype)
            matcher_x = get_match_fun(x, patterntype)
            pred = lambda n: matcher_m(n) and not matcher_x(n)

    return pred


def findfolder(toppath, match='*', exclude=''):
    """
    recursively find folder path from toppath.
    patterns to decide to walk folder path or not
    :type toppath: str
    :type match: str or list(str)
    :type exclude: str or list(str)
    :rtype: generator for path str
    """
    pred = _pred_pattern(match, exclude)

    return (p for p in walkfolder(toppath, pred))


def walkfolder(toppath, pred):
    """
    walk folder if pred(foldername) is True
    :type toppath: str
    :type pred: function(str) => bool
    """
    for entry in scandir.scandir(toppath):
        if not entry.is_dir() or not pred(entry.name):
            continue
        yield entry.path
        for p in walkfolder(entry.path, pred):
            yield p


def tempdir():
    import tempfile
    d = tempfile.gettempdir()
    p = os.path.join(d, 'her_temp')
    mkdir_if_not(p, ispath=True)

    return p


def tempfolder(prefix=''):
    """임시 폴더를 만들어서 리턴"""
    import uuid

    p = prefix + str(uuid.uuid4())
    d = tempdir()
    tmpd = os.path.join(d, p)
    return mkdir_if_not(tmpd, ispath=True)


def tempfile(mode, ext='', **kwargs):
    import uuid

    d = tempdir()

    if ext and not ext.startswith('.'):
        ext = '.' + ext

    fname = os.path.join(d, str(uuid.uuid4()) + ext)
    return open(fname, mode, **kwargs)


def renderhtml(template, **kwargs):
    # from packageutil import caller_path
    from .caller import caller

    from jinja2 import Environment, FileSystemLoader
    from os.path import dirname

    if '/' not in template:
        p = dirname(caller.abspath(depth=2))
    else:
        p, template = os.path.split(template)

    j2_env = Environment(loader=FileSystemLoader(p),
                         trim_blocks=True)
    rendered = j2_env.get_template(template).render(**kwargs)
    return rendered


def renderimages(images, width=80, height=80, space=0):
    import webbrowser

    template = os.path.dirname(os.path.realpath(__file__)) + '/template/images.html'
    rendered = renderhtml(template, data=images, width=width, height=height, space=space)
    tmp = tempfile('wt', '.html')
    tmp.write(rendered)
    tmp.flush()

    webbrowser.open(tmp.name)
    return tmp.name


def imsize(fname):
    """
    return image size (height, width)
    :param fname:
    :return:
    """
    from PIL import Image
    im = Image.open(fname)
    return im.size[1], im.size[0]


def imread(fname, size=None, expand=True, dtype='float32', **kwargs):
    from skimage import io, transform
    import numpy as np

    img = io.imread(fname, **kwargs)
    image_max = np.iinfo(img.dtype).max
    if size is not None:
        sz = list(img.shape)
        sz[:len(size)] = size
        img = transform.resize(img, sz, preserve_range=True)
    if dtype.startswith('float'):
        # normalize 0 to 1
        img = img.astype(dtype) / float(image_max)
    else:
        img = img.astype(dtype)

    if expand:
        img = np.expand_dims(img, 0)
        if img.ndim == 3:
            img = np.expand_dims(img, -1)
    return img


def imsave(fname, *args, **kwargs):
    from skimage import io

    mkdir_if_not(fname)
    res = io.imsave(fname, *args, **kwargs)
    logg.info('image saved to [{}]'.format(fname))

    return res


def imread_palette(fname, expand=True, dtype='uint8', mode='r'):
    from PIL import Image

    # use png? https://pythonhosted.org/pypng/png.html ?

    img = Image.open(fname, mode=mode)
    palette = img.getpalette()
    if palette is not None:
        img = np.array(img)
        num_colors = len(palette) / 3

        image_max = float(np.iinfo(img.dtype).max)
        palette = np.array(palette).reshape(num_colors, 3) / image_max

    else:
        colors = img.convert('RGBA').getcolors()
        num_colors = len(colors)
        assert num_colors <= 256
        palette = [c[:3] for _, c in colors]
        im = Image.new('P', img.size)
        palette = np.array(palette).reshape((-1))
        im.putpalette(palette, rawmode="RGB")
        im.paste(img)
        palette = im.getpalette()
        img = np.array(im)

    img = img.astype(dtype)
    palette = palette.astype('float32')
    if expand:
        img = np.expand_dims(img, axis=0)
        palette = np.expand_dims(palette, axis=0)

    return img, palette


def filecopy(src, dst):
    import shutil
    shutil.copy(src, dst)


def download(url, out=None):
    import wget
    if out is not None:
        mkdir_if_not(out)
    logg.info('downloading... [{}] to [{}]'.format(url, out))
    f = wget.download(url, out=out)

    return f


def download_if_not(url, f):
    if not os.path.exists(f):
        f = download(url, f)
    return f


def unzip(z, member=None):
    import zipfile
    path = os.path.dirname(z)
    zip = zipfile.ZipFile(z, 'r')
    if member is None:
        zip.extractall(path)
        logg.info('unzip [{}] to [{}]'.format(z, path))
    else:
        zip.extract(member, path)
        logg.info('unzip [{}] to [{}]'.format(member, path))
    zip.close()


def untar(t, member=None):
    import tarfile
    path = os.path.dirname(t)
    tar = tarfile.open(t)
    if member is None:
        tar.extractall(path)
        logg.info('unzip [{}] to [{}]'.format(t, path))
    else:
        tar.extract(member, path)
        logg.info('unzip [{}] to [{}]'.format(member, path))
    tar.close()


def anyfile(pattern):
    import glob
    return any(glob.glob(pattern))



# -*- coding: utf-8 -*-
from functools import wraps
from contextlib import contextmanager
import signal


def functional(ifunctional):
    """
    fun(fn) -> function or
    fun(fn, args...) -> call of fn(args...)
    :param ifunctional: f
    :return: decorated function
    """

    @wraps(ifunctional)
    def wrapper(fn, *args, **kw):

        fn = ifunctional(fn)
        if args or kw:
            return fn(*args, **kw)
        else:
            return fn

    return wrapper


def tuple_pack(*args):
    return args


@functional
def tupleout(fn):
    """

    ex) code ::

        def asis(a):
        return a

        f = tupleout(asis)
        print f(1)  # (1,)
        print tupleout(asis, 1)  # (1,)

    :param Function fn:
    :return:
    """
    @wraps(fn)
    def wrapped(*args, **kwargs):
        return fn(*args, **kwargs),

    return wrapped


def tuplefy(x):
    """
    make tuple if not tuple or list
    :param x:
    :return:
    """
    if x is None or isinstance(x, (tuple, list)):
        return x
    else:
        return x,


def tuple_arg(fn):
    """
    fun(1,2) -> fun((1,), (2,))로
    f(1,2,3) =>  f((1,), (2,), (3,))
    :param fn:
    :return:
    """
    @wraps(fn)
    def wrapped(*args, **kwargs):
        args = map(tuplefy, args)
        return fn(*args, **kwargs)

    return wrapped


def tuple_args(fn):
    """
    args 파싱 유틸 function
    fun(p1, p2, ...pn, **kwargs) or fun([p1, p2, ..], **kwargs)
    ex) 샘플::

        @tuple_arg
        def f(args, **kwargs):
            for d in args:
                print d
        f(1,2,3) =>  f([1,2,3])
    :param function fn:
    :return:
    """

    @wraps(fn)
    def wrapped(*args, **kwargs):

        if len(args) == 1:
            if isinstance(args[0], tuple):
                return fn(args[0], **kwargs)
            elif isinstance(args[0], list):
                return fn(tuple(args[0]), **kwargs)
        return fn(args, **kwargs)

    return wrapped


def unpack_args(classfun, nth=0):
    """
    args 갯수가 nth + 1 개 일때, 그게 만약 tuple이면, unpack
    :param classfun:
    :param nth: nth = 0, 일반 함수, nth = 1: 클래스 함수 1이 self니깐.
    :return:
    """
    if classfun:
        nth = 1

    def deco(fn):
        def wrapped(*args, **kwargs):
            if len(args) == nth + 1 and isinstance(args[nth], (tuple, list)):
                args = tuple(args[:nth] + args[nth])
                return fn(*args, **kwargs)
            else:
                return fn(*args, **kwargs)
        return wrapped

    return deco


def optional_str(deco):
    """
    string 1개만 deco 인자로 오거나 없거나.
    :param deco:
    :return:
    """

    @wraps(deco)
    def dispatcher(*args, **kwargs):

        # when only function arg
        if not kwargs and len(args) == 1 and not isinstance(args[0], str) \
                and args[0] is not None:
            decorator = deco()
            return decorator(args[0])
        else:
            # decorator with args
            decorator = deco(*args, **kwargs)
            return decorator

    return dispatcher


def patchmethod(*cls, **kwargs):
    """
    클래스 멤버 패치 @patchmethod(Cls1, ..., [name='membername'])
    ex)
    class A(object):
        def __init__(self, data):
            self.data = data

    @patchmethod(A)
    def sample(self):
        ''' haha docstrings '''
        print self.data

    @patchmethod(A, name='membermethod)
    def sample(self):
        ''' haha docstrings '''
        print self.data

    a = A()
    a.sample()

    """

    def _patch(fun):
        m = kwargs.pop('name', None) or fun.__name__
        for c in cls:
            setattr(c, m, fun)
            # c.__dict__[m].__doc__ = fun.__doc__

    def wrap(fun):
        _patch(fun)
        return fun

    return wrap


def patchproperty(*cls, **kwargs):
    """
    class getter 함수 패치 decorator
    EX)
    class B(A):
        pass

    @patchproperty(B)
    def prop(self):
        return 'hello'

    :param cls:
    :param kwargs:
    """

    def _patch(fun):
        m = kwargs.pop('property', None) or fun.__name__
        p = property(fun)
        for c in cls:
            setattr(c, m, p)

    def wrap(fun):
        _patch(fun)
        return fun

    return wrap


class patch(object):

    @staticmethod
    def method(clz, fun, name=None):
        name = name or fun.__name__
        for c in clz:
            setattr(c, name, fun)

    @staticmethod
    def methods(clz, funs, names=None):
        if names is None:
            for f in funs:
                name = f.__name__
                for c in clz:
                    setattr(c, name, f)
        else:
            for f, name in zip(funs, names):
                for c in clz:
                    setattr(c, name, f)

    @staticmethod
    def getter(clz, fun, name=None):
        name = name or fun.__name__
        p = property(fun)
        for c in clz:
            setattr(c, name, p)

    @staticmethod
    def getters(clz, funs):
        for f in funs:
            p = property(f)
            name = f.__name__
            for c in clz:
                setattr(c, name, p)


@contextmanager
def on_interrupt(handler, reraise=False):
    """
    context for handling keyboardinterrupt
    ex)
    with on_interrupt(handler):
        critical_work_to_prevent()

    from logger import logg
    on_interrupt.signal = None

    :param function handler:
    :param bool reraise:
    :return: context
    """

    def _handler(sig, frame):
        handler.signal = (sig, frame)
        handler._reraise = handler()

    handler._reraise = False
    handler.signal = None
    oldhandler = signal.getsignal(signal.SIGINT)
    signal.signal(signal.SIGINT, _handler)

    yield handler

    signal.signal(signal.SIGINT, oldhandler)
    if (reraise or handler._reraise) and handler.signal:
        oldhandler(*handler.signal)


def interrupt_guard(msg='', reraise=True):
    """
    context for guard keyboardinterrupt
    ex)
    with interrupt_guard('need long time'):
        critical_work_to_prevent()

    :param str msg: message to print when interrupted
    :param reraise: re-raise or not when exit
    :return: context
    """
    def echo():
        print(msg)

    return on_interrupt(echo, reraise=reraise)





# -*- coding: utf-8 -*-
from .queues import QueueInterruptable, is_main_alive
from .concurrent import Threaded

from .ilogging import logg
from time import sleep


class ActiveQ(QueueInterruptable):

    def __init__(self, maxsize=256, jobq=None, resq=None):
        self._run_thread = None
        self._resq = resq  # result q
        self._jobq = jobq  # input q
        super(ActiveQ, self).__init__(maxsize)

    def start(self):
        # assert self._thread is None
        assert self._run_thread is None
        self._run_thread = Threaded(self._run)
        return self

    def _run(self):
        logg.info('run started')

        while not self.finish and is_main_alive():
            try:
                item = self.get(block=False, timeout=0.01)
            except QueueInterruptable.Empty as e:
                # raise e
                sleep(0.01)
                if self._jobq:
                    try:
                        item = self._jobq.get(block=False)
                    except StopIteration:
                        logg.info('stop iteration')
                        break
                    except QueueInterruptable.Empty as e:
                        continue
                    self.put(item, block=True)
                continue
            # logg.info(item)
            assert isinstance(item, tuple)

            # fun, args, kwargs = item
            # res = fun(*args, **kwargs)
            res = self.action(item)

            if self._resq:
                self._resq.put(res, block=True)

        self._run_thread = None

    def action(self, item):
        """
        for overriding
        :param item:
        :return:
        """
        fun, args, kwargs = item
        return fun(*args, **kwargs)

    def push_job(self, fun, *args, **kwargs):
        """
        put job if possible, non-blocking
        :param fun:
        :param args:
        :param kwargs:
        :return:
        """
        assert callable(fun)
        return self.put((fun, args, kwargs), block=True)

    def put_job(self, fun, *args, **kwargs):
        """
        put job if possible, non-blocking
        :param fun:
        :param args:
        :param kwargs:
        :return:
        """
        if not args and not kwargs and isinstance(fun, (tuple, list)):
            # ex) q.put_job([fun, args, kwargs])
            fun, args, kwargs = fun

        assert callable(fun)
        return self.put((fun, args, kwargs), block=False)


if __name__ == '__main__':
    import sys

    def testfun(i):
        print(i)
        sys.stdout.flush()
        return i

    q = ActiveQ(10).start()

    for i in range(10):
        q.push_job(testfun, i)

    while not q.empty():
        pass

    q.stop()
    print('done')



import numpy as np
from snipy import irandom as rand
from snipy.basic import tuple_args
# from snipy.concurrent import SharedPool

# todo imread with resize ..
# todo : add some examples

# region image augmentation oriented
# assume : images [B,H,W,C] formats


# region rand util


def rand_apply_batch(fun, imagez, **kwargs):
    assert isinstance(imagez, (tuple, list))
    # print imagez[0].shape, imagez[1].shape
    # print len(imagez[0])
    mask = np.random.choice([False, True], len(imagez[0]), p=kwargs.pop('p', None))
    # print 'masks', mask

    def apply(images):
        data = fun(images[mask], **kwargs)
        images[mask] = data
        return images

    return tuple(map(apply, imagez)) if np.any(mask) else imagez


def rand_apply_onebatch(fun, imagez, **kwargs):
    assert isinstance(imagez, (tuple, list))
    mask = np.random.choice([False, True], len(imagez[0]), p=kwargs.pop('p', None))
    # if mask is True, apply function to tuple of images. ex: [image, labelimage]

    def apply(images):
        images[mask] = [fun(d, **kwargs) for d in images[mask]]
        return images

    return tuple(map(apply, imagez)) if np.any(mask) else imagez

# endregion

# region random augmentation


def rand_augmentations(level=3):
    funs = (rand_fliplr, rand_flipud, rand_rot90,
            rand_brightness, rand_intensity, rand_elastic,)
    return funs[:level]


def rand_fliplr(*imagez):
    """ flip together """
    return rand_apply_onebatch(np.fliplr, imagez)


def rand_flipud(*imagez):
    """ flip together """
    return rand_apply_onebatch(np.flipud, imagez)


def rand_rot90(*imagez):
    """ rotate together """
    return rand_apply_onebatch(np.rot90, imagez)


def rand_crop(sz, *imagez):
    """
    random crop
    # assume imagez has same size (H, W)
    # assume sz is less or equal than size of image
    :param sz: cropped image sz
    :param imagez: imagez
    :return: rand cropped image pairs or function bound to sz
    """

    def _rand_crop(*imgz):
        imsz = imgz[0].shape[:2]

        assert imsz[0] >= sz[0] and imsz[1] >= sz[1]

        si = np.random.randint(imsz[0] - sz[0]) if imsz[0] > sz[0] else 0
        sj = np.random.randint(imsz[1] - sz[1]) if imsz[1] > sz[1] else 0

        slicei = slice(si, si+sz[0])
        slicej = slice(sj, sj+sz[1])

        outs = tuple(img[slicei, slicej] for img in imgz)
        return tuple_or_not(*outs)

    return _rand_crop(*imagez) if imagez else _rand_crop


def tuple_or_not(*args):
    return args[0] if len(args) == 1 else args


def rand_rotate(anglerange, *imagez):
    """
    :param anglerange:
    :param imagez:
    :return:
    """
    r = float(anglerange[1] - anglerange[0])
    s = anglerange[0]

    def _rand_rotate(*imgz):
        angle = np.random.random(1)[0] * r + s
        out = tuple(rotate(img, angle) for img in imgz)
        return tuple_or_not(out)

    return _rand_rotate(*imagez) if imagez else _rand_rotate


def rand_elastic(*imagez, **kwargs):
    return rand_apply_onebatch(elastic_transform, imagez, **kwargs)


def rand_blend(images):
    # make blending mask
    # apply to imagez
    inoise = rand_blend_mask(images.shape)
    discrete = np.round(inoise * images.shape[3])
    return blend_discrete(images, discrete)


def blend_discrete(images, depthmask, depth=None):
    """
    depthmask : shape of [batch, h, w]
    """
    imshape = images.shape
    depth = depth or images.shape[3]
    blend = np.empty(shape=(imshape[0], imshape[1], imshape[2]))
    for d in range(depth):
        imask = (depthmask == d)
        channel = images[..., d]
        blend[imask] = channel[imask]
    return np.expand_dims(blend, axis=-1)


def rand_blend_mask(shape, rand=rand.uniform(-10, 10), **kwargs):
    """ random blending masks """
    # batch, channel = shape[0], shape[3]
    z = rand(shape[0])  # seed
    noise = snoise2dz((shape[1], shape[2]), z, **kwargs)

    return noise  # np.round(noise * channel)


def snoise2dvec(size, *params, **kwargs):  #, vlacunarity):
    """
    vector parameters
    :param size:
    :param vz:
    :param vscale:
    :param voctave:
    :param vpersistence:
    :param vlacunarity:
    :return:
    """
    data = (snoise2d(size, *p, **kwargs) for p in zip(*params))  # , vlacunarity))
    return np.stack(data, 0)


def snoise2d(size, z=0.0, scale=0.05, octaves=1, persistence=0.25, lacunarity=2.0):
    """
    z value as like a seed
    """
    import noise
    data = np.empty(size, dtype='float32')
    for y in range(size[0]):
        for x in range(size[1]):
            v = noise.snoise3(x * scale, y * scale, z,
                              octaves=octaves, persistence=persistence, lacunarity=lacunarity)
            data[x, y] = v
    data = data * 0.5 + 0.5
    if __debug__:
        assert data.min() >= 0. and data.max() <= 1.0
    return data


def snoise2dz(size, z, scale=0.05, octaves=1, persistence=0.25, lacunarity=2.0):
    """
    z as seeds
    scale이 작을 수록 패턴이 커지는 효과
    """
    import noise
    z_l = len(z)

    data = np.empty((z_l, size[0], size[1]), dtype='float32')
    for iz in range(z_l):
        zvalue = z[iz]
        for y in range(size[0]):
            for x in range(size[1]):
                v = noise.snoise3(x * scale, y * scale, zvalue,
                                  octaves=octaves, persistence=persistence, lacunarity=lacunarity)
                data[iz, y, x] = v
    data = data * 0.5 + 0.5
    if __debug__:
        assert data.min() >= 0. and data.max() <= 1.0
    return data


@tuple_args
def rand_brightness(imagez, scale=1.0, randfun=rand.normal(0., .1), clamp=(0., 1.)):
    """
    :param images:
    :param scale: scale for random value
    :param randfun: any randfun binding except shape
    :param clamp: clamping range
    :return:
    """
    l, h = clamp
    r = randfun((imagez[0].shape[0], 1, 1, 1)) * scale

    def apply(im):
        im += r
        im[im < l] = l
        im[im > h] = h
        return im

    return tuple(map(apply, imagez))


import warnings
np.seterr(all='raise')

@tuple_args
def rand_intensity(imagez, scale=1.0, randfun=rand.normal(1., 0.1), clamp=(0., 1.)):

    l, h = clamp
    # c = (l + h) / 2.
    bshape = (imagez[0].shape[0], 1, 1, 1)
    phi = randfun(bshape) * scale
    theta = randfun(bshape) * scale
    p = randfun(bshape) * scale

    def apply(im):
        # (h / phi) * (im/(h/theta))**p
        with warnings.catch_warnings():
            warnings.filterwarnings('error')
            try:
                im2 = (h * phi) * np.power(im * theta / h, p)
                im2[im2 < l] = l
                im2[im2 > h] = h
            except Exception as e:
                # logg.warn('in rand_intensity: [%s], pass' % e)
                return im
        # im *= r
        return im2

    return tuple(map(apply, imagez))


# endregion

# region image transform implementations
def elastic_transform(im, alpha=0.5, sigma=0.2, affine_sigma=1.):
    """
    Based on https://gist.github.com/erniejunior/601cdf56d2b424757de5
    elastic deformation of images as described in [Simard2003]
    """
    # fixme : not implemented for multi channel !
    import cv2

    islist = isinstance(im, (tuple, list))
    ima = im[0] if islist else im

    # image shape
    shape = ima.shape
    shape_size = shape[:2]

    # Random affine transform
    center_square = np.float32(shape_size) // 2
    square_size = min(shape_size) // 3
    pts1 = np.float32([center_square + square_size,
                       [center_square[0] + square_size, center_square[1] - square_size],
                       center_square - square_size])
    pts2 = pts1 + np.random.uniform(-affine_sigma, affine_sigma, size=pts1.shape).astype(np.float32)
    M = cv2.getAffineTransform(pts1, pts2)
    if islist:
        res = []
        for i, ima in enumerate(im):
            if i == 0:
                res.append(cv2.warpAffine(ima, M, shape_size[::-1], borderMode=cv2.BORDER_REFLECT_101))
            else:
                res.append(cv2.warpAffine(ima, M, shape_size[::-1]))
        im = res
    else:
        ima = cv2.warpAffine(ima, M, shape_size[::-1], borderMode=cv2.BORDER_REFLECT_101)
        # ima = cv2.warpAffine(ima, M, shape_size[::-1])

    # fast gaussian filter
    blur_size = int(4 * sigma) | 1
    dx = cv2.GaussianBlur((np.random.rand(*shape) * 2 - 1), ksize=(blur_size, blur_size), sigmaX=sigma) * alpha
    dy = cv2.GaussianBlur((np.random.rand(*shape) * 2 - 1), ksize=(blur_size, blur_size), sigmaX=sigma) * alpha

    # remap
    x, y = np.meshgrid(np.arange(shape[1]), np.arange(shape[0]))
    map_x, map_y = (y + dy).astype('float32'), (x + dx).astype('float32')

    def remap(data):
        r = cv2.remap(data, map_y, map_x, interpolation=cv2.INTER_LINEAR, borderMode=cv2.BORDER_REFLECT_101)
        return r[..., np.newaxis]

    if islist:
        return tuple([remap(ima) for ima in im])
    else:
        return remap(ima)

# endregion


def rotate_crop(centerij, sz, angle, img=None, mode='constant', **kwargs):
    """
    rotate and crop
    if no img, then return crop function
    :param centerij:
    :param sz:
    :param angle:
    :param img: [h,w,d]
    :param mode: padding option
    :return: cropped image or function
    """
    # crop enough size ( 2 * sqrt(sum(sz^2) )
    # rotate
    from skimage import transform
    sz = np.array(sz)
    crop_half = int(np.ceil(np.sqrt(np.square(sz).sum())))

    if centerij[0] >= crop_half or centerij[1] >= crop_half:
        raise NotImplementedError

    slicei = slice(centerij[0] - crop_half, centerij[0] + crop_half)
    slicej = slice(centerij[1] - crop_half, centerij[1] + crop_half)

    # slicei = (centerij[0] - crop_half, centerij[0] + crop_half)
    # slicej = (centerij[1] - crop_half, centerij[1] + crop_half)

    # def _pad_if_need(im):
    #     imshape = im.shape
    #     pad_need = slicei[0] < 0 or slicej[0] < 0 or slice
    #     padwidth = [(slicei[0], np.maximum(0, slicei[1] - imshape[0])),
    #                 (slicej[0], np.maximum(0, slicej[1] - imshape[1]))]

    def _rotate_cropcenter(im):
        enoughcrop = im[slicei, slicej]

        rotated = transform.rotate(enoughcrop, angle, resize=False, preserve_range=True, mode=mode, **kwargs)
        return cropcenter(sz, rotated)
    if img is not None:
        return _rotate_cropcenter(img)

    return _rotate_cropcenter


def rotate(img, angle, resize=False, preserve_range=True, mode='constant', **kwargs):
    from skimage import transform
    return transform.rotate(img, angle, resize=resize, preserve_range=preserve_range, mode=mode, **kwargs)


def crop(img, center, sz, mode='constant'):
    """
    crop sz from ij as center
    :param img:
    :param center: ij
    :param sz:
    :param mode:
    :return:
    """
    center = np.array(center)
    sz = np.array(sz)
    istart = (center - sz / 2.).astype('int32')
    iend = istart + sz
    imsz = img.shape[:2]
    if np.any(istart < 0) or np.any(iend > imsz):
        # padding
        padwidth = [(np.minimum(0, istart[0]), np.maximum(0, iend[0]-imsz[0])),
                    (np.minimum(0, istart[1]), np.maximum(0, iend[1]-imsz[1]))]
        padwidth += [(0, 0)] * (len(img.shape) - 2)
        img = np.pad(img, padwidth, mode=mode)
        istart = (np.maximum(0, istart[0]), np.maximum(0, istart[1]))

        return img[istart[0]:istart[0]+sz[0], istart[1]:istart[1]+sz[1]]

    return img[istart[0]:iend[0], istart[1]:iend[1]]


def cropcenter(sz, img=None):
    """
    if no img, then return crop function
    :param sz:
    :param img:
    :return:
    """
    l = len(sz)
    sz = np.array(sz)

    def wrapped(im):
        imsz = np.array(im.shape)
        s = (imsz[:l] - sz) / 2  # start index
        to = s + sz  # end index

        # img[s[0]:to[0], ... s[end]:to[end], ...]
        slices = [slice(s, e) for s, e in zip(s, to)]

        return im[slices]

    if img is not None:
        return wrapped(img)

    return wrapped


def cropcenter_batch(sz, img=None):
    return cropcenter_dim(sz, start_axis=1, img=img)


def cropcenter_dim(sz, start_axis=0, img=None):
    l = len(sz)
    sz = np.array(sz)

    def wrapped(im):
        imsz = np.array(im.shape)
        s = (imsz[start_axis:l+start_axis] - sz) / 2
        to = s + sz
        # img[:, s[0]:to[0], ... s[end]:to[end], ...]
        slices = [slice(None)] * start_axis + [slice(s, e) for s, e in zip(s, to)]
        return im[slices]

    if img is not None:
        return wrapped(img)

    return wrapped


def pad_if_need(sz_atleast, img, mode='constant'):
    # fixme : function or ....
    """
    pad img if need to guarantee minumum size
    :param sz_atleast: [H,W] at least
    :param img: image np.array [H,W, ...]
    :param mode: str, padding mode
    :return: padded image or asis if enought size
    """
    # sz_atleast = np.asarray(sz_atleast)
    imsz = img.shape[:2]  # assume img [H,W, ...]
    padneed = np.asarray((sz_atleast[0] - imsz[0], sz_atleast[1] - imsz[1]))
    if np.any(padneed > 0):
        # need padding
        padding = np.zeros((img.ndim, 2), dtype='int16')
        padneed = np.maximum(padneed, 0)
        padding[:2, 0] = padneed/2
        padding[:2, 1] = padneed - padneed/2
        img = np.pad(img, padding, mode=mode)

    return img


def canny(img, threshold1=255/3, threshold2=255, **kwargs):
    """ canny edge """
    import cv2
    # edges=None, apertureSize=None, L2gradient=None
    if img.ndim <= 3:
        edge = cv2.Canny(img, threshold1, threshold2, **kwargs)
        if edge.ndim == 2:
            edge = np.expand_dims(edge, 2)
    elif img.ndim == 4:
        # batch
        edge = np.asarray([cv2.Canny(i, threshold1, threshold2, **kwargs) for i in img])
        if edge.ndim == 3:
            edge = np.expand_dims(edge, 3)
    else:
        raise ValueError('above 5d?')
    return edge


def _convert_uint8(im):
    if im.dtype != np.uint8:
        im = np.uint8(im * 255)
    return im


@tuple_args
def alpha_composite(images):
    from PIL import Image

    # images

    # backward
    # images = reversed(images)
    # out = Image
    out = Image.fromarray(_convert_uint8(images[0]))
    for img in images[1:]:
        img = Image.fromarray(_convert_uint8(img))
        # check same size?
        # Image.alpha_composite(bg, fg)
        # alpha composite im2 over im1
        out = Image.alpha_composite(img, out)

    return np.asarray(out, dtype='float32') / 255.




# -*- coding: utf-8 -*-
from collections import (OrderedDict, MutableMapping, Mapping)


class ODict(MutableMapping):

    def __init__(self, *args, **kwargs):
        self._odict = OrderedDict(*args, **kwargs)

    def __setitem__(self, key, value):
        self._odict[key] = value

    def __getitem__(self, item):
        return self._odict[item]

    def __delitem__(self, key):
        del self.__dict__[key]

    def __setattr__(self, key, value):
        if key == '_odict':
            super(ODict, self).__setattr__(key, value)
        else:
            self._odict[key] = value

    def __getattr__(self, item):
        return self._odict.get(item, None)

    def __iter__(self):
        return iter(self._odict)

    def __len__(self):
        return len(self._odict)

    def __str__(self):
        return str(self._odict).replace('OrderedDict', 'odict', count=1)

    def __repr__(self):
        return self._odict.__repr__().replace('OrderedDict', 'odict', count=1)

    def pop(self, key, **kw):
        return self._odict.pop(key, **kw)

    def popitem(self):
        return self._odict.popitem()

    def clear(self):
        self._odict.clear()

    def update(self, *args, **kwds):
        self._odict.update(*args, **kwds)

    def setdefault(self, key, default=None):
        return self._odict.setdefault(key, default=default)

    def __sub__(self, other):
        return ODict((k, v) for k, v in self.items() if k not in other)

    def __add__(self, other):
        if not other:
            return self.copy()
        added = self.copy()
        added.update(other)
        return added

    __radd__ = __add__

    def __getnewargs__(self):
        """pickling related"""
        return tuple()

    def __getstate__(self):
        return self._odict

    def __setstate__(self, d):
        self._odict.update(d)

    def copy(self):
        return ODict((k, v) for k, v in self.items())

    def intersect(self, other):
        """
        self와 other 키가 동일한 아이템의 dictobj
        :type other: dict
        :rtype: dictobj:
        """
        return ODict((k, self[k]) for k in self if k in other)

    @staticmethod
    def from_dict(dic):
        """
        recursive dict to dictobj 컨버트
        :param dic:
        :return:
        """
        return ODict((k, ODict.convert_ifdic(v)) for k, v in dic.items())

    @staticmethod
    def convert_ifdic(value):
        if isinstance(value, Mapping):
            return ODict.from_dict(value)
        else:
            return value


# alias
odict = ODict




# -*- coding: utf-8 -*-
import time
import sys
import math
from .dictobj import dictobj
from .term import write, clear, TermCursor


class ProgressBase(object):

    defaultopt = dictobj(show=True, update_per=1, show_item=False, width=10)

    def __init__(self, iter=None, **kwargs):
        self.iter = iter
        self.opt = self.defaultopt.copy()
        self.update_opt(iter, kwargs)
        write('\n')

    def update_opt(self, iter, args=None):
        self.iter = iter or self.iter
        if args:
            args = {k: v for k, v in args.items() if v is not None}
            self.opt.update(args)
        self.opt.start = time.time()
        self.opt.seen_so_far = 0
        self.setup()

    @property
    def target(self):
        return self.opt.target

    @target.setter
    def target(self, value):
        self.opt.target = value
        self.setup()

    def setup(self):
        pass

    def __call__(self, iter=None, **kwargs):
        self.update_opt(iter, kwargs)
        return self

    def __iter__(self):
        if not self.opt.show:
            return iter(self.iter)
        else:
            return self.iter_verbose(enum=False)

    def enumerate(self, iter=None, **kwargs):
        self.update_opt(iter, kwargs)
        if not self.opt.show:
            return enumerate(self.iter)
        else:
            return self.iter_verbose(enum=True)

    def iter_verbose(self, enum=False):

        def enum_yied(ind, data):
            return ind, data

        def plain_yield(_, data):
            return data

        fun = enum_yied if enum else plain_yield
        update_per = self.opt.update_per
        show_item = self.opt.show_item
        for i, item in enumerate(self.iter):
            if i % update_per == 0:
                if show_item:
                    self.update(i, item)
                else:
                    self.update(i)
            yield fun(i, item)
        self.stop()

    def stop(self, message=None):
        self.write(message or '[Done]')

    @classmethod
    def write(cls, message):
        if message:
            sys.stdout.write(str(message))
        sys.stdout.flush()

    def format_current(self, current):
        bar = self.format_bar(current)
        bar += self.format_eta(1.0)
        return bar

    def update(self, current, message=None):
        bar = self.format_current(current)
        write(clear.line + TermCursor.home)
        # write(TermCursor.home)
        if message:
            self.write(bar + str(message))
        else:
            self.write(bar)

    def format_bar(self, current):
        bar = '[%d]' % (current,)
        width = self.opt.width
        bar += self.format_progress(current % width, width, width)
        return bar

    def format_eta(self, ratio):
        from .ansi import ansi
        if ratio == 0:
            return ansi.green('-[ETA:unknown]')
        elif ratio < 1.:
            return ansi.green('-[ETA:{0:.2f}h]'.format(self.get_eta(ratio)))
        else:
            return ansi.green('-[ETA:{0:.2f}h]'.format((time.time() - self.opt.start)/3600.))

    def get_eta(self, ratio):
        elapsed = time.time() - self.opt.start
        eta = 0 if ratio == 0 else elapsed / ratio * (1 - ratio)

        return eta/3600.

    @classmethod
    def format_progress(cls, current, target, width):
        from .ansi import ansi
        prog = int(float(current) / target * width)
        arrow = '>' if current < target else '='
        bar = '=' * prog + arrow + '.' * (width - prog)

        return ansi.yellow('[' + bar + ']')


# noinspection PyAttributeOutsideInit
class Progress(ProgressBase):

    defaultopt = dictobj(target=None, width=10, show=True, step=None, update_per=None,
                         show_item=False)

    def __init__(self, iter=None, **kwargs):
        super(Progress, self).__init__(iter, **kwargs)

    def setup(self):
        super(Progress, self).setup()

        opt = self.opt

        if opt.target is None:
            target = len(self.iter) if self.iter else opt.target
            target = target or opt.width
        else:
            target = opt.target

        step = opt.step or target
        self.opt.target = target
        opt.update_per = opt.update_per or target / float(step)
        opt.bar_format = self.bar_format(target)
        # opt.seen_so_far = 0

    def stop(self, message=None):
        self.update(self.target, message)

    def format_current(self, current):
        bar = self.format_bar(current)
        bar += self.format_eta(float(current)/self.target)
        return bar

    @classmethod
    def bar_format(cls, target):
        if target is None:
            return ' %s/%s'
        numdigits = int(math.floor(math.log10(target))) + 1
        barstr = '  %{0}d/%{0}d '.format(numdigits)
        return barstr

    def format_bar(self, current):
        opt = self.opt
        bar = opt.bar_format % (current, self.target)
        bar += self.format_progress(current, self.target, opt.width)
        return bar


def progress(iter, **kwargs):
    """
    프로그래스 bar
    for i in progress(10):
        print i

    for i in progress(iter):
        print i
    """
    if isinstance(iter, int):
        iter = xrange(iter)
    if hasattr(iter, '__len__') or 'target' in kwargs:
        cls = Progress
    else:
        cls = ProgressBase

    return cls(iter, **kwargs)


def iprogress(iter, **kwargs):
    """
    enumerate with progress
    """
    return progress(iter, **kwargs).enumerate()




# -*- coding: utf-8 -*-
from contextlib import closing
import MySQLdb
import MySQLdb.cursors
from _mysql_exceptions import OperationalError
from .basic import patch


def db_config(db):
    """
    :param db: str: 디비 스키마
    :return: mysql connection 공통 옵션
    """
    return {
        'host': '127.0.0.1',
        'user': "default_user",
        'passwd': "default_pass",
        'charset': 'utf8',
        'db': db
    }


def database(db='', **kwargs):
    """
    usage:
    with database('my_db') as conn:
        c = conn.cursor()
        ....
    database 커넥션 with 문과 같이 사용하고, 알아서 close하기
    :param db: str: db스키마
    :param kwargs:
    :return:
    """
    db = kwargs.pop('db', db)
    arg = db_config(db)
    arg.update(kwargs)

    return closing(MySQLdb.connect(**arg))


def connect(db='', **kwargs):
    """
    db 접속 공통 인자들 채워서 접속, schema만 넣으면 됩니다.
    db connection 객체 반환이지만
    with 문과 같이 쓰이면 cursor임에 주의 (MySQLdb의 구현이 그렇습니다.)
    ex1)
    import snipy.database as db
    conn = db.connect('my_db')
    cursor = conn.cursor()

    ex2)
    import snipy.database as db
    with db.connect('my_db') as cursor:
        cursor.execute(query)

    :param db: str: db schema
    :param kwargs: 추가 접속 정보
    :return: connection or cursor
    """
    arg = db_config(db)
    arg.update(kwargs)
    return MySQLdb.connect(**arg)


def _cursor_exit(cursor, exc_type, exc_value, traceback):
    """
    cursor with문과 쓸수 있게 __exit__에 바인딩
    :param cursor:
    :param exc_type:
    :param exc_value:
    :param traceback:
    :return:
    """
    if exc_type is not None:
        print(exc_value, traceback)
    cursor.connection.close()


def cursor(db, **kwargs):
    """
    db cursor 리턴
    :param db: str: db 스키마 이름
    :param kwargs: 추가 MySQLdb.connect 의 kwarg들
    :return: cursor 오브젝트
    """
    return connect(db, **kwargs).cursor()

# 커서를 with문과 쓸수 있게 patch
setattr(MySQLdb.cursors.Cursor, '__enter__', lambda cur: cur)
setattr(MySQLdb.cursors.Cursor, '__exit__', _cursor_exit)


def execute(db, query, args=None, **kwargs):
    c = cursor(db, **kwargs)
    # logging.info(query)
    try:
        c.execute(query, args=args)
    except OperationalError as e:
        print(query, args)
        print(e.message, e.args)
        raise e
    return c


def fetch(query, args=None, **kwargs):
    """
    for record in fetch(query, args, **configs):
        print record
    :param args:
    :param db: str: db 스키마
    :param query: 쿼리 스트링
    :param kwargs: db connection 추가 인자. 보통 생략
    :return: iterator
    """
    cur = execute(kwargs.pop('db', ''), query, args, **kwargs)

    for r in cur:
        yield r

    cur.connection.close()

# """
# cursor method patching
#
# """


def fieldcount(cursor):
    return len(cursor.description)


def fieldnames(cursor):
    """
    최근 실행한 쿼리의 field names 리스트
    :param cursor:
    :return: list[str]
    """
    return [f[0] for f in cursor.description]


def commit(cursor):
    cursor.connection.commit()


def get_insert_query(table, fields=None, field_count=None):
    """
    format insert query
    :param table: str
    :param fields: list[str]
    :param field_count: int
    :return: str
    """
    if fields:
        q = 'insert into %s ({0}) values ({1});' % table
        l = len(fields)
        q = q.format(','.join(fields), ','.join(['%s'] * l))
    elif field_count:
        q = 'insert into %s values ({0});' % table
        q = q.format(','.join(['%s'] * field_count))
    else:
        raise ValueError('fields or field_count need')

    return q


# noinspection PyShadowingNames
def insert(cursor, table, *args, **field_values):
    """
    db에 레코드 집어넣기
    ex)
    cursor.insert(table, v1, v2,...)
    ex)
    cursor.insert(table, id=v1, word=v2, commit=True)
    :param commit:
    :param cursor:
    :param table:
    :param args:
    :param field_values:
    :return:
    """
    commit = field_values.pop('commit', True)

    q, a = None, None
    if args is not None and len(args) > 0:
        q = get_insert_query(table, field_count=len(args))
        a = args
    elif len(field_values) > 0:
        q = get_insert_query(table, fields=field_values.keys())
        a = field_values.values()
    else:
        raise ValueError('need table, record...')

    cursor.execute(q, args=a)
    if commit:
        cursor.connection.commit()


# noinspection PyShadowingNames
def update(cursor, table, where_kv, commit=True, **field_values):
    """
    db update 쿼리 빌딩 및 실행, 단, commit은
    :param cursor: 커서
    :type cursor: Cursor
    :param table: 테이블 이름
    :type table: str
    :param where_kv: 업데이트 where 조건 dictionary, key:field, value:equal condition only
    :type where_kv: dict
    :param field_values: kwarg 업데이트용
    :type field_values: dict
    :param commit: 커밋 여부
    :type commit: bool
    :return:
    """

    q = """update %s \nset {0} \nwhere {1} """ % table

    fields = field_values.keys()
    kv = ','.join(['{}=%s'.format(f) for f in fields])

    where = ' and '.join(['{}=%s'.format(f) for f in where_kv.keys()])

    q = q.format(kv, where)
    args = field_values.values() + where_kv.values()

    cursor.execute(q, args=args)
    if commit:
        cursor.connection.commit()


# noinspection PyShadowingNames
def insert_or_update(cursor, table, commit=True, **field_values):
    """
    db update 쿼리 빌딩 및 실행, 단, commit은
    :param cursor: 커서
    :type cursor: Cursor
    :param table: 테이블이름
    :type table: str
    :param commit: 커밋 여부
    :type commit: bool
    :param field_values: insert 또는 업데이트 할 필드 및 값 dict pairs
    :type field_values:dict
    :return:
    """

    q = """INSERT INTO %s ({0}) \nVALUES ({1}) \nON DUPLICATE KEY UPDATE {2} """ % table
    l = len(field_values)

    fields = field_values.keys()
    field = ','.join(fields)
    value = ','.join(['%s'] * l)
    kv = ','.join(['{}=%s'.format(f) for f in fields])

    q = q.format(field, value, kv)
    args = field_values.values() * 2

    cursor.execute(q, args=args)
    if commit:
        cursor.connection.commit()

# add property
MySQLdb.cursors.Cursor.fieldcount = property(fieldcount)

# cursor method patching
patch.methods((MySQLdb.cursors.Cursor,), [commit, fieldnames, insert, update, insert_or_update])




# -*- coding: utf-8 -*-
"""
caller, stack info
"""
import os
import inspect
import sys


class caller(object):

    """
    caller stack inspection, 호출하는 함수 입장에서 caller
    caller.path()
    caller.abspath()
    caller.modulename()
    caller.funname()
    ex) def p():
            # funname of calling p()
            callername = caller.funname()

    """

    extra_depth = 1

    @classmethod
    def path(cls, depth=1):
        """ caller path (*.py)
        """
        depth += cls.extra_depth
        return inspect.stack()[depth][1]

    @classmethod
    def abspath(cls, depth=1):
        """ caller path (*.py)
        """
        return os.path.abspath(cls.path(depth + 1))

    @classmethod
    def funname(cls, depth=1):
        depth += cls.extra_depth
        return inspect.stack()[depth][3]

    @classmethod
    def modulename(cls, depth=1):
        """
        get caller's __name__
        """
        depth += cls.extra_depth
        frame = sys._getframe(depth)
        return frame.f_globals['__name__']

    @classmethod
    def packagename(cls, depth=1):
        """
        get package name by splitting __name__
        """
        name = cls.modulename(depth+1)
        return name.split('.', 1)[0]


class this(caller):

    """
    this stack inspection, 호출하는 곳에서.
    this.path()
    this.abspath()
    this.modulename()
    this.funname()
    ex) def p():
            assert 'p' == this.funname()
    """
    extra_depth = 0




# -*- coding: utf-8 -*-
"""
package util, inspection related
"""
import os
import sys
from os.path import dirname

# fixme
from .ilogging import logg


def guess_package_path(searchfrom):
    """
    package path. return None if failed to guess
    """
    from snipy.io import fileutil

    current = searchfrom + '/'
    init_found = False
    pack_found = False
    while not init_found and current != '/':
        current = os.path.dirname(current)
        initfile = os.path.join(current, '__init__.py')
        init_found = os.path.exists(initfile)

    if not init_found:
        # search for breadth
        searchfrom = dirname(searchfrom)

        for folder in fileutil.listfolder(searchfrom):
            current = os.path.join(searchfrom, folder)
            initfile = os.path.join(current, '__init__.py')
            init_found = os.path.exists(initfile)
            if init_found:
                break

    while init_found:
        current = os.path.dirname(current)
        initfile = os.path.join(current, '__init__.py')
        init_found = os.path.exists(initfile)
        pack_found = not init_found

    return current if pack_found else None


def find_package_path(searchfrom):
    """
    package path. return None if failed to guess
    """

    current = searchfrom + '/'
    init_found = False
    pack_found = False
    while not init_found and current != '/':
        current = os.path.dirname(current)
        initfile = os.path.join(current, '__init__.py')
        init_found = os.path.exists(initfile)

    while init_found:
        current = os.path.dirname(current)
        initfile = os.path.join(current, '__init__.py')
        init_found = os.path.exists(initfile)
        pack_found = not init_found

    return current if pack_found else None


def append_sys_path(p):
    """
    append system path
    """
    if p not in sys.path:
        sys.path.insert(0, p)


def append_this_package_path(depth=1):
    """
    this_package.py 에서 사용
    import snipy.this_package
    """
    from .caller import caller

    logg.debug('caller module %s', caller.modulename(depth + 1))
    c = caller.abspath(depth + 1)
    logg.debug('caller path %s', c)

    p = guess_package_path(dirname(c))
    if p:
        logg.debug('appending sys path %s', p)
        append_sys_path(p)
    else:
        # do some logging
        logg.debug('failed to guess package path for: %s', c)


def import_this(packagename):
    if packagename in sys.modules:
        return
    import importlib
    append_this_package_path(depth=1)
    importlib.import_module(packagename)




# -*- coding: utf-8 -*-
import functools
from .basic import tuple_args

from .concurrent import Threaded
from .listx import ListCallable, ListArgs


def get_call_chain(something):
    try:
        return something.get_call_chain()
    except AttributeError:
        return [something]


class IterChain(object):
    def __init__(self, left, right):
        self._iterables = get_call_chain(left) + get_call_chain(right)

    def get_call_chain(self):
        return self._iterables

    def __call__(self, it):
        chained = it
        for i in self._iterables:
            chained = i(chained)
        return chained

    def __iter__(self):
        chained = self._iterables[0]
        for i in self._iterables[1:]:
            chained = i(chained)
        return chained

    def __rshift__(self, right):
        return IterChain(self, right)

    def __rrshift__(self, left):
        return IterChain(left, self)

    __lshift__ = __rrshift__
    __rlshift__ = __rshift__

    def __str__(self):
        return 'IterChain%s' % str(self._iterables)


class iterable(object):
    """
    decorator
    ex)
    @iterable
    def example(n):
        for i in range(n):
            yield i
    e1 = example(10)
    e2 = example(8)
    for i in e1:
        print i
    for i in e2:
        print i
    """

    def __init__(self, gen):
        self._gen = gen

    def __call__(self, *args, **kwargs):
        return iterator(self._gen, *args, **kwargs)


class iterator(object):
    """
    reiterable object
    """

    def __init__(self, gen, *args, **kwargs):
        import threading

        self.gen = gen
        self.it = None
        self.args = args  # ()
        self.kwargs = kwargs  # {}
        self.lock = threading.Lock()

        functools.update_wrapper(self, gen)

        if args or kwargs:
            self.reset()

    def __call__(self, *args, **kwargs):
        """
        decorator call
        """
        self.args = args
        self.kwargs = kwargs
        self.reset()
        return self

    def __iter__(self):
        if self.it is None:
            self.reset()
        return self

    def next(self):
        # http://anandology.com/blog/using-iterators-and-generators/
        with self.lock:
            if self.it is None:
                raise StopIteration
            try:
                return self.it.next()
            except StopIteration:
                self.it = None
                raise StopIteration

    def reset(self):
        self.it = self.gen(*self.args, **self.kwargs)

    def __str__(self):
        return str(self.gen)

    def __repr__(self):
        return str(self.gen)


def iterate(g, *args, **kwargs):
    it = iterator(g)
    return it(*args, **kwargs)


class iterflow(iterable):

    def __rshift__(self, other):
        return IterChain(self, other)

    def __rrshift__(self, other):
        return IterChain(other, self)

    def __lshift__(self, other):
        return IterChain(other, self)

    def __rlshift__(self, other):
        return IterChain(self, other)

    def get_call_chain(self):
        return [self]


class FlowList(ListArgs):

    def __iter__(self):
        res = self[0]
        for call in self[1:]:
            res = call(res)
        return res

    def __call__(self, *args, **kwargs):
        res = self[0](*args, **kwargs)
        for call in self[1:]:
            res = call(res)
        return res


@tuple_args
def flows(args):
    """
    todo : add some example
    :param args:
    :return:
    """

    def flow_if_not(fun):
        # t = type(fun)
        if isinstance(fun, iterator):
            return fun
        elif isinstance(fun, type) and 'itertools' in str(fun.__class__):
            return fun
        else:
            try:
                return flow(fun)
            except AttributeError:
                # generator object has no attribute '__module__'
                return fun

    return FlowList(map(flow_if_not, args))


def flow(fun):

    @iterflow
    @functools.wraps(fun)
    def _gen(it):
        for data in it:
            yield fun(data)

    return _gen


def iflow(ifun):

    @iterflow
    @functools.wraps(ifun)
    def _gen(it):
        for data in it:
            for d in ifun(data):
                yield d
    return _gen


def applyit(f):

    @flow
    @functools.wraps(f)
    def apply(data):
        return [f(i) for i in data]

    return apply


def loop(times=None):
    """
    None for infinite loop
    :param times:
    :return:
    """
    import itertools
    return iterate(itertools.repeat, True, times)


@iterflow
def forever(it):
    """ forever
    todo : add example
    """
    while True:
        # generator 두번쨰 iteration 무한 루프 방지
        i = iter(it)
        try:
            yield i.next()
        except StopIteration:
            raise StopIteration
        while True:
            try:
                yield i.next()
            except StopIteration:
                break


@iterflow
def irange(*args, **kwargs):
    for i in xrange(*args, **kwargs):
        yield i


def nexts(count, iterable=None):

    @iterflow
    def nextsit(it):
        for _ in loop(count):
            yield it.next()

    return nextsit(iterable) if iterable else nextsit


def cycle(c, iterable=None):

    @iterflow
    def cycleit(it):
        for _ in loop(c):
            i = iter(it)
            for d in i:
                yield d

    return cycleit(iterable) if iterable else cycleit


def ifilter(pred=None, iterable=None):
    pred = pred or bool

    @iterflow
    def _filter(it):
        for d in it:
            if pred(d):
                yield d

    return _filter(iterable) if iterable else _filter


@flow
def zipflow(data):
    return zip(*data)


def ibatch(size, iterable=None, rest=False):
    """
    add example
    :param size:
    :param iterable:
    :param rest:
    :return:
    """

    @iterflow
    def exact_size(it):
        it = iter(it)
        while True:
            yield [it.next() for _ in xrange(size)]

    @iterflow
    def at_most(it):
        it = iter(it)
        while True:
            data = []
            for _ in xrange(size):
                try:
                    data.append(it.next())
                except StopIteration:
                    if data:
                        yield data
                    raise StopIteration
            yield data

    ibatchit = at_most if rest else exact_size

    return ibatchit if iterable is None else ibatchit(iterable)


def batchzip(size, iterable=None, rest=False):
    """
    todo : add example
    :param size:
    :param iterable:
    :param rest:
    :return:
    """
    fn = ibatch(size, rest=rest) >> zipflow

    return fn if iterable is None else fn(iterable)


# todo check nargout == 1 or .. some refactoring
def batchstack(size, iterable=None, rest=False):
    """
    todo : add example
    :param size:
    :param iterable:
    :param rest:
    :return:
    """

    def stack(data):
        import numpy as np
        return map(np.vstack, data)

    fn = batchzip(size, rest=rest) >> flow(stack)

    return fn if iterable is None else fn(iterable)


def shuffle(qsize=1024, iterable=None):
    """
    add example
    :param qsize:
    :param iterable:
    :return:
    """

    @iterflow
    def shuffleit(it):
        from random import randrange
        q = []

        for i, d in enumerate(it):
            q.insert(randrange(0, len(q) + 1), d)
            if i < qsize:
                continue
            yield q.pop(randrange(0, len(q)))

        while q:
            yield q.pop(randrange(0, len(q)))

    return shuffleit if iterable is None else shuffleit(iterable)


def queue_class(QueueClass):

    class _iqueue(QueueClass, iterflow):

        __name__ = 'iqueue'

        def __init__(self, iterable=None, maxsize=1024, processor=None):
            super(_iqueue, self).__init__(maxsize=maxsize)
            self.producer = iterable
            self.proc = processor
            self.th = None
            if iterable:
                self.produce()

        def produce(self):

            # self.th = Thread(target=self._produce, args=(self.producer,))
            # self.th.start()
            self.th = Threaded(self._produce, self.producer)

        def _produce(self, producer):
            if self.proc:
                for d in producer:
                    # self.put(self.proc(d))
                    if self.put(self.proc(d)) is False:
                        break

            else:
                for d in producer:
                    # self.put(d)
                    if self.put(d) is False:
                        break

        def __call__(self, iterable):
            self.producer = iterable
            self.produce()
            return self

        def __iter__(self):
            if self.done:
                self.reset()
            return self

        def next(self):
            while not self.done:
                try:
                    return self.get(block=False, timeout=0.01)
                except QueueClass.Empty:
                    continue
            else:
                raise StopIteration

            # if not self.done:
            #     return self.get(block=True)
            # else:
            #     raise StopIteration

        def reset(self):
            # reset = produce
            print('reset Q')
            self.produce()

        @property
        def done(self):
            if self.th is None:
                return True
            return self.empty() and not self.th.is_alive()

        def __repr__(self):
            return str(self.producer)

    return _iqueue


# multiprocessing queue has sometime unexpectable bloken pipe ... T T
# todo(dade) test with pyspark
from .queues import QueueInterruptable

_IQueueMulti = queue_class(QueueInterruptable)


def iqueue(iterable=None, maxsize=1024, processor=None, qclass=None):
    if qclass is None:
        return _IQueueMulti(iterable, maxsize, processor)
    else:
        qklass = queue_class(qclass)
        return qklass(iterable, maxsize, processor)





# -*- coding: utf-8 -*-
from snipy.ilogging import logg
from snipy.basic import tuple_args
from snipy.progress import progress
import contextlib
try:
    import matplotlib.pyplot as plt
except ImportError as e:
    # apt update && apt install -y python3-tk.  if need ploting module
    print('warning: apt update && apt install -y python3-tk. if need ploting module', e)
    pass

import numpy as np


@tuple_args
def plots(data, **kwargs):
    """
    simple wrapper plot with labels and skip x
    :param yonly_or_xy:
    :param kwargs:
    :return:
    """
    labels = kwargs.pop('labels', '')
    loc = kwargs.pop('loc', 1)

    # if len(yonly_or_xy) == 1:
    #     x = range(len(yonly_or_xy))
    #     y = yonly_or_xy
    # else:
    #     x = yonly_or_xy[0]
    #     y = yonly_or_xy[1:]

    lines = plt.plot(np.asarray(data).T, **kwargs)
    if labels:
        plt.legend(lines, labels, loc=loc)
    return lines


def grid_recommend(count, ratio=(2., 3.)):
    # ratio = (30, 40)
    unit = np.sqrt(count / float(np.prod(ratio)))
    h = np.ceil(ratio[0] * unit)
    w = np.ceil(count / float(h))
    return int(h), int(w)


def imshow_grid(images, grid=None, showfun=None, **opt):
    """
    :param images: nhwc
    :return:
    """
    # assert images.ndim == 4  or list
    showfun = showfun or plt.imshow
    count = len(images)
    grid = grid or grid_recommend(count, sorted(images[0].shape[:2]))

    res = []
    for i, img in enumerate(images):
        # grid row first index
        plt.subplot2grid(grid, (i % grid[0], i // grid[0]))
        res.append(showfun(img.squeeze(), **opt))

    return res

# todo enhance. with enumerate


def plt_range(*args, **kwargs):
    """
    for i in plot_range(n):
        plt.imshow(imgs[i])

    left arrow yield prev value
    other key yield next value
    :param args:
    :return:
    """
    wait = kwargs.pop('wait', True)
    if not wait:
        # no interactive just pass range
        for i in progress(range(*args)):
            yield i
        return

    class _holder(object):
        pass
    hold = _holder()
    hold.i = 0
    hold.done = False

    def press(event):
        # import sys
        # sys.stdout.flush()
        hold.i += -1 if event.key == 'left' else 1
        hold.i = 0 if hold.i < 0 else hold.i

    def onclose(event):
        hold.done = True

    fig = kwargs.pop('fig', None)
    figsize = kwargs.pop('figsize', None)
    if fig is None:
        fig = plt.gcf()
        if figsize:
            fig.set_size_inches(figsize)
    elif isinstance(fig, (int, str)):
        if figsize:
            fig = plt.figure(fig, figsize=figsize)
        else:
            fig = plt.figure(fig)
    elif isinstance(fig, plt.Figure):
        if figsize:
            fig.set_size_inches(figsize)
    else:
        raise ValueError

    onkey_fig(press, fig)
    onclose_fig(onclose, fig)

    ranges = range(*args)
    l = len(ranges)

    while hold.i < l:
        print('hold.i', ranges[hold.i])
        yield ranges[hold.i]  # yield first without keypress
        before = hold.i
        while before == hold.i:
            while not fig.waitforbuttonpress(0.01):
                if hold.done:
                    return
            while fig.waitforbuttonpress(0.1):
                if hold.done:
                    return


def pause(interval):
    return plt.pause(interval)


def plot_pause(timeout=None, msg=''):
    """
    todo : add some example
    :param timeout: wait time. if None, blocking
    :param msg:
    :return:
    """

    if timeout is not None:
        print(msg or 'Press key for continue in time {}'.format(timeout))
        plt.waitforbuttonpress(timeout=timeout)
        return True

    print(msg or 'Press key for continue')
    while not plt.waitforbuttonpress(timeout=0.01):
        if not plt.get_fignums():
            return False
    return len(plt.get_fignums()) != 0


def ploting(iterable, timeout=None, msg=''):
    for res in iterable:
        yield res
        if not plot_pause(timeout, msg):
            break
    raise StopIteration


def flat_images(images, grid=None, bfill=1.0, bsz=(1, 1)):
    """
    convert batch image to flat image with margin inserted
    [B,h,w,c] => [H,W,c]
    :param images:
    :param grid: patch grid cell size of (Row, Col)
    :param bfill: board filling value
    :param bsz: int or (int, int) board size
    :return: flatted image
    """
    if images.ndim == 4 and images.shape[-1] == 1:
        images = images.squeeze(axis=-1)

    grid = grid or grid_recommend(len(images), sorted(images[0].shape[:2]))
    if not isinstance(bsz, (tuple, list)):
        bsz = (bsz, bsz)

    # np.empty()
    imshape = list(images.shape)
    imshape[0] = grid[0] * grid[1]
    imshape[1] += bsz[0]
    imshape[2] += bsz[1]

    # data = np.empty((grid[0] * grid[1], imshape[1], imshape[2]), dtype=images.dtype)
    data = np.empty(imshape, dtype=images.dtype)

    data.fill(bfill)
    bslice0 = slice(0, -bsz[0]) if bsz[0] else slice(None, None)
    bslice1 = slice(0, -bsz[1]) if bsz[1] else slice(None, None)

    data[:len(images), bslice0, bslice1] = images

    imshape = list(grid) + imshape[1:]  # [grid[0], grid[1], H, W, [Channel]]
    data = data.reshape(imshape)
    if len(imshape) == 5:
        data = data.transpose(0, 2, 1, 3, 4)
        imshape = [imshape[0]*imshape[2], imshape[1]*imshape[3], imshape[4]]
    else:  # len == 4
        data = data.transpose(0, 2, 1, 3)
        imshape = [imshape[0]*imshape[2], imshape[1]*imshape[3]]
    data = data.reshape(imshape)

    # remove last margin
    data = data[bslice0, bslice1]

    return data


def imshow_flat(images, grid=None, showfun=None, bfill=1.0, bsz=(1,1), **opt):
    """
    imshow after applying flat_images
    :param images: [bhwc]
    :param grid: None for auto grid
    :param showfun: plt.imshow
    :param bfill: color for board fill
    :param bsz: size of board
    :param opt: option for showfun
    :return:
    """
    showfun = showfun or plt.imshow

    count = len(images)
    # decide grid shape if need pick one
    grid = grid or grid_recommend(count, ratio=sorted(images[0].shape[:2]))

    flatted = flat_images(images, grid, bfill=bfill, bsz=bsz)
    res = showfun(flatted, **opt)
    plt.draw()


def imshows_flat(images, subgrid=None, grid=None, showfun=None, bfill=1.0, bsz=(1,1), **opt):
    showfun = showfun or plt.imshow
    count = len(images[0])

    grid = grid or grid_recommend(count, ratio=sorted(images[0][0].shape[:2]))
    flatted = [flat_images(im, grid, bfill=bfill, bsz=bsz) for im in images]
    subgrid = subgrid or (1, len(images))
    return imshow_grid(flatted, grid=subgrid, showfun=showfun, **opt)


@tuple_args
def imshow(imagez, grid=None, showfun=None, bfill=1.0, bsz=(1,1), tight=True, **opt):
    showfun = showfun or plt.imshow
    if len(imagez) == 1:
        if imagez[0].ndim == 2:
            res = showfun(imagez[0], **opt)
            plt.draw()
        elif imagez[0].ndim == 3:
            res = showfun(imagez[0], **opt)
            plt.draw()
        else:
            res = imshow_flat(imagez[0], grid=grid, showfun=showfun, bfill=bfill, bsz=bsz, **opt)
    else:
        # note : imshows_flat
        res = imshows_flat(imagez, grid=grid, showfun=showfun, bfill=bfill, bsz=bsz, **opt)
    if tight:
        plt.tight_layout()
    return res


def onclose_fig(onclose, fig=None):
    fig = fig or plt.gcf()
    fig.canvas.mpl_connect('close_event', onclose)


def onkey_fig(onkey, fig=None):
    fig = fig or plt.gcf()
    fig.canvas.mpl_connect('key_press_event', onkey)


def matshow(*args, **kwargs):
    """
    imshow without interpolation like as matshow
    :param args:
    :param kwargs:
    :return:
    """
    kwargs['interpolation'] = kwargs.pop('interpolation', 'none')
    return plt.imshow(*args, **kwargs)

# region plot loop tool


class PlotLoop(object):
    """
    imshow grid for looping speed up
    http://stackoverflow.com/questions/33602185/speed-up-plotting-images-in-matplotlib

    Example:

        import sflow.python.ploting as plt
        plot = plt.plotloop(plt.imshow_flat)
        for i in range(nstep):
            if i % 50 == 0:
                out = sess.run([styled, trainop])[0]
                plot(out)
                plt.pause(0.0001)
                print(i)
            else:
                l = sess.run([loss_content, loss_style, trainop])[:-1]
                print(i, l)
        plt.plot_pause()

    """

    def __init__(self, showfun=imshow, fig=None, **drawopt):
        self.showfun = showfun
        self.fig = fig or plt.figure()
        self.drawopt = drawopt
        self.setdata = None
        self.onclose = drawopt.pop('onclose', self._onclose)

    def _onclose(self, evt):
        # import sys
        # print('figure closed, exit')
        # sys.exit(0)
        pass

    def __call__(self, *args, **kwargs):

        if not self.setdata:
            if self.fig is None:
                self.fig = plt.figure()
            self.fig.canvas.mpl_connect('close_event', self.onclose)
            kwargs.update(self.drawopt)
            plt.figure(self.fig.number)
            self.setdata = self.showfun(*args, **kwargs)

            plt.show(block=False)
        else:
            # todo@dade : fixme
            try:
                self.setdata.set_data(*args, **kwargs)
            except AttributeError:
                plt.figure(self.fig.number)
                self.showfun(*args, **kwargs)

            self.fig.canvas.draw()
        plt.pause(0.0001)


class PlotGrid(object):
    """
    imshow grid for looping speed up
    http://stackoverflow.com/questions/33602185/speed-up-plotting-images-in-matplotlib

    """

    def __init__(self, grid=None, ratio=(2., 4.), showfun=imshow, fig=None, onclose=None, **drawopt):
        self.showfun = showfun
        self.ratio = ratio
        self.grid = grid
        self.setdata = []
        # self.figsize = drawopt.pop('figsize', None)
        self.drawopt = drawopt
        self.fig = fig or plt.figure()
        self.onclose = onclose or self._onclose

    def _onclose(self, evt):
        # import sys
        # print('figure closed, exit')
        # sys.exit(0)
        pass

    def __call__(self, images, **kwargs):
        grid = self.grid or grid_recommend(len(images), kwargs.pop('ratio', self.ratio))

        if not self.setdata:
            if self.fig is None:
                self.fig = plt.figure()
            self.fig.canvas.mpl_connect('close_event', self.onclose)
            kwargs.update(self.drawopt)
            for i, img in enumerate(images):
                plt.subplot2grid(grid, (i % grid[0], i // grid[0]))
                self.setdata.append(self.showfun(img.squeeze(), **kwargs))

        else:
            # todo@dade : fixme
            for i, img in enumerate(images):
                self.setdata[i].set_data(img.squeeze())
            # except AttributeError:
            #     for i, img in enumerate(images):
            #         plt.subplot2grid(grid, (i % grid[0], i // grid[0]))
            #         self.showfun(img.squeeze(), **kwargs)

            self.fig.canvas.draw()


plotloop = PlotLoop
plotgrid = PlotGrid


#


# region overlay and objects


def imshow_overlay(im1, im2, alpha, **kwargs):
    plt.imshow(im1)
    plt.hold(True)
    return plt.imshow(im2, alpha=alpha, **kwargs)


def imbox(xy, w, h, angle=0.0, **kwargs):
    """
    draw boundary box
    :param xy: start index xy (ji)
    :param w: width
    :param h: height
    :param angle:
    :param kwargs:
    :return:
    """
    from matplotlib.patches import Rectangle
    return imbound(Rectangle, xy, w, h, angle, **kwargs)


def imrect(xy, w, h, angle=0.0, **kwargs):
    # http://matplotlib.org/api/patches_api.html#matplotlib.patches.Rectangle
    from matplotlib.patches import Rectangle
    return impatch(Rectangle, xy, w, h, angle, **kwargs)


def imbound(clspatch, *args, **kwargs):
    """
    :param clspatch:
    :param args:
    :param kwargs:
    :return:
    """
    # todo : add example

    c = kwargs.pop('color', kwargs.get('edgecolor', None))
    kwargs.update(facecolor='none', edgecolor=c)
    return impatch(clspatch, *args, **kwargs)


def impatch(clspatch, *args, **kwargs):
    # http://matplotlib.org/api/patches_api.html#matplotlib.patches.Rectangle
    from matplotlib.patches import Rectangle
    ax = plt.gca()
    clspatch = clspatch or Rectangle
    patch = clspatch(*args, **kwargs)
    ax.add_patch(patch)

    return patch


def bar(values, labels=None, txtcolor='white', fmt='%s', txtoffset=-0.05, **kwargs):
    x = range(len(values))
    res = plt.bar(x, values)
    ax = plt.gca()
    if labels is not None:
        plt.xticks(np.asarray(x) + 0.4, labels)

    for ix, v in zip(x, values):
        ax.text(ix + 0.4, v + txtoffset, fmt % v, color=txtcolor, horizontalalignment='center',
                weight='bold', **kwargs)

    return res, ax

# endregion


# region superpixel

def imslic(img, n_segments=100, aspect=None):
    """
    slic args :
    n_segments=100, compactness=10., max_iter=10,
    sigma=0, spacing=None,
    multichannel=True, convert2lab=None, enforce_connectivity=True,
    min_size_factor=0.5, max_size_factor=3, slic_zero=False

    mark_boundaries args:
    label_img, color=(1, 1, 0), outline_color=None, mode='outer', background_label=0

    imshow args:
    cmap=None, norm=None, aspect=None, interpolation=None,
    alpha=None, vmin=None, vmax=None, origin=None,
    extent=None, shape=None, filternorm=1, filterrad=4.0,
    imlim=None, resample=None, url=None, hold=None, data=None,

    :param img:
    :param slicarg:
    :param slickw:
    :return:
    """
    from skimage.segmentation import (slic, mark_boundaries)
    from skimage.morphology import (dilation)

    if img.ndim == 2 or img.ndim == 3 and img.shape[-1] == 1:
        imz = np.stack([img, img, img], 2)
    else:
        imz = img

    slics = slic(imz, n_segments=n_segments)

    boundaries = mark_boundaries(imz, slics)
    return plt.imshow(boundaries, aspect=aspect)


def imslic2(img, n_segments=100, color=None, outline_color=None, mode='thick', **kwargs):
    """
    slic args :
    n_segments=100, compactness=10., max_iter=10,
    sigma=0, spacing=None,
    multichannel=True, convert2lab=None, enforce_connectivity=True,
    min_size_factor=0.5, max_size_factor=3, slic_zero=False

    mark_boundaries args:
    label_img, color=(1, 1, 0), outline_color=None, mode='outer', background_label=0

    imshow args:
    cmap=None, norm=None, aspect=None, interpolation=None,
    alpha=None, vmin=None, vmax=None, origin=None,
    extent=None, shape=None, filternorm=1, filterrad=4.0,
    imlim=None, resample=None, url=None, hold=None, data=None,

    :param img:
    :param slicarg:
    :param slickw:
    :return:
    """
    from skimage.segmentation import (slic, find_boundaries) # mark_boundaries
    from skimage.morphology import (dilation)

    kwslic = {'compactness', 'max_iter', 'sigma', 'spacing', 'multichannel', 'convert2lab',
              'enforce_connectivity', 'min_size_factor', 'max_size_factor', 'slic_zero=False'}
    imshowkw = {'cmap', 'norm', 'aspect', 'interpolation', 'alpha', 'vmin', 'vmax', 'origin',
                'extent', 'shape', 'filternorm', 'filterrad', 'imlim', 'resample', 'url', 'hold', 'data'}

    slicarg = {k: v for k, v in kwargs.iteritems() if k in kwslic}
    imshowarg = {k: v for k, v in kwargs.iteritems() if k in imshowkw}

    if img.ndim == 2 or img.ndim == 3 and img.shape[-1] == 1:
        imz = np.stack([img, img, img], 2)
        color = color or 1.
    else:
        imgz = img
        color = color or (1,1,0)

    slics = slic(imz, n_segments=n_segments, **slicarg)

    boundaries = find_boundaries(slics, mode=mode)
    if outline_color is not None:
        outlines = dilation(boundaries, np.ones((3, 3), np.uint8))
        img[outlines] = outline_color
    img[boundaries] = color
    return plt.imshow(img, **imshowarg)

# endregion

# region movie

# from matplotlib.animation import (FFMpegWriter, ImageMagickFileWriter, AVConvWriter,
#                                   FFMpegFileWriter, ImageMagickWriter, AVConvFileWriter)
from matplotlib.animation import (FFMpegWriter, ImageMagickWriter, AVConvWriter)


class PlotMovieWriter(object):
    """
    PlotLoop + MovieWriter
    Example:
        import sflow.python.ploting as plt
        plot = plt.PlotMovieWriter(plt.imshow_flat, outputfile, dpi=100)
        for i in range(nstep):
            if i % 50 == 0:
                out = sess.run([styled, trainop])[0]
                plot(out)
                # plt.pause(0.0001)
                print(i)
            else:
                l = sess.run([loss_content, loss_style, trainop])[:-1]
                print(i, l)
        plot.finish()

        plt.plot_pause()

    """

    def __init__(self, outfile, showfun=None, fig=None, drawopt=None, dpi=100, **movieopt):

        self.showfun = showfun or plt.imshow
        self.fig = fig or plt.figure()
        drawopt = drawopt or dict()
        self.drawopt = drawopt
        self.setdata = None
        self.onclose = drawopt.pop('onclose', self._onclose)
        # for movie writing
        self.moviewriter = None
        self.movieopt = movieopt

        self.outfile = outfile
        self.dpi = dpi
        self._first = True

    def setup_movie(self, fig):
        # create moviewriter
        # then setup
        # self.movieopt.pop()

        # fps=5, codec=None, bitrate=None, extra_args=None, metadata
        self.moviewriter = FFMpegWriter(**self.movieopt)
        self.moviewriter.setup(fig, self.outfile, self.dpi)

    def _onclose(self, evt):
        # import sys
        # print('figure closed, exit')
        # sys.exit(0)
        pass

    def __call__(self, *args, **kwargs):

        if not self.setdata:
            if self.fig is None:
                self.fig = plt.figure()
                # self.fig = self.setup_figure()
                self.fig.canvas.mpl_connect('close_event', self.onclose)
            kwargs.update(self.drawopt)
            self.setdata = self.showfun(*args, **kwargs)
            if self._first:
                self.setup_movie(self.fig)
                self._first = False

            plt.show(block=False)
        else:
            # todo@dade : if error? showfun?
            try:
                self.setdata.set_data(*args, **kwargs)
            except AttributeError:
                self.showfun(*args, **kwargs)

            self.fig.canvas.draw()

        self.grab()

    def grab(self):
        self.moviewriter.grab_frame()

    def finish(self):
        if self._first:
            raise ValueError('setup not called')
        self.moviewriter.finish()
        logg.info('movie saved to [{}]'.format(self.outfile))


def get_ax_size(fig, ax):
    bbox = ax.get_window_extent()
    x_pad = ax.xaxis.get_tick_padding()
    y_pad = ax.yaxis.get_tick_padding()
    bbox.x0 += x_pad
    bbox.x1 -= x_pad
    bbox.y0 += y_pad
    bbox.y1 -= y_pad
    bbox = bbox.transformed(fig.dpi_scale_trans.inverted())
    width, height = bbox.width, bbox.height
    # width *= fig.dpi
    # height *= fig.dpi
    return width, height


def tight_figure(fig):
    a = fig.gca()
    a.set_position([0, 0, 1, 1])
    a.set_xticks([])
    a.set_yticks([])

    # a.xaxis.set_tick_params(which='both', pad=0, reset=True)
    # a.yaxis.set_tick_params(which='both', pad=0, reset=True)
    plt.axis('off')
    fig.patch.set_visible(False)
    # force update and get axsize
    plt.show(False)
    w, h = get_ax_size(fig, a)
    fig.set_size_inches(w, h, forward=True)


class ImageMovieWriter(PlotMovieWriter):

    def __init__(self, outfile, *args, **kwargs):
        super(ImageMovieWriter, self).__init__(outfile, *args, **kwargs)

    def setup_movie(self, fig):
        # remove boarders
        tight_figure(fig)
        super(ImageMovieWriter, self).setup_movie(fig)


@contextlib.contextmanager
def movie_saving(outfile, showfun=imshow, fig=None, tight=True, drawopt=None, dpi=100, **movieopt):
    """
    contextmanager for PlotMovieWriter
    Example:

        with movie_saving('output.mp4', dpi=100) as plot:
            for i in range(10):
                plot(data[i])

    :param outfile:
    :param showfun:
    :param fig:
    :param tight:
    :param drawopt:
    :param dpi:
    :param movieopt: fps=5, codec=None, bitrate=None, extra_args=None, metadata=None
    :return:
    """
    if tight:
        plot_writer = ImageMovieWriter(outfile, showfun=showfun, fig=fig, drawopt=drawopt, dpi=dpi, **movieopt)
    else:
        plot_writer = PlotMovieWriter(outfile, showfun=showfun, fig=fig, drawopt=drawopt, dpi=dpi, **movieopt)

    try:
        yield plot_writer
    finally:
        plot_writer.finish()


# endregion


# if __name__ == '__main__':
#     import numpy as np
#     a = np.random.rand(4, 3, 5, 3)
#     b = flat_images(a, (2, 2), 0., (1, 1))
#     # plot_pause()
#     print('done')
#



# -*- coding: utf-8 -*-

"""
https://en.wikipedia.org/wiki/ANSI_escape_code
http://wiki.bash-hackers.org/scripting/terminalcodes
http://ascii-table.com/ansi-escape-sequences.php
http://www.ecma-international.org/publications/files/ECMA-ST/Ecma-048.pdf
"""
import termios
import sys
# from dictobj import dictobj

try:
    # prepare terminal settings
    _fd = sys.stdin.fileno()
    _old_settings = termios.tcgetattr(_fd)
    _new_settings = termios.tcgetattr(_fd)
    _new_settings[3] &= ~termios.ICANON
    _new_settings[3] &= ~termios.ECHO

except Exception as e:
    # from .ilogging import logg
    # logg.warn('not a tty?: %s:' % e)
    pass


def echo(content):
    write(str(content))


def write(content):
    """
    write and flush terminal code and(or) text
    """
    sys.stdout.write(content)
    sys.stdout.flush()


def csi_fixed(code):
    return '\x1b[%s' % code


class csi(object):
    """
    ansi csi code simple abstraction
    """
    def __init__(self, code):
        self.code = '\x1b[%s' + code

    def __call__(self, n=''):
        return self.code % str(n)

    def __str__(self):
        return self.__call__()


class csi_move(csi):
    """
    csi move command
    """
    def __call__(self, x='', y=''):
        xy = '%s;%s' % (str(x), str(y))
        return self.code % xy


class TermCursor(object):
    """
    ansi codes for TermCursor movement
    """

    up = csi('A')
    down = csi('B')
    right = csi('C')
    left = csi('D')
    newline = downhome = csi('E')
    uphome = csi('F')
    movex = csi('G')
    move = csi_move('H')
    # home = csi('E')() + csi('A')()  # downhome(1) + up(1)
    home = '\r'
    save = csi_fixed('s')
    restore = csi_fixed('u')
    hide = csi_fixed('?25l')
    show = csi_fixed('?25h')

    @staticmethod
    def moveby(x, y):
        fx = TermCursor.right(x) if x > 0 else TermCursor.left(-x)
        fy = TermCursor.down(y) if y > 0 else TermCursor.up(-y)
        return fx + fy


def save_pos():
    write(TermCursor.save)


def restore_pos():
    write(TermCursor.restore)


class clear(object):
    """
    ansi codes for clearing terminal
    """
    screen_after = csi_fixed('J')
    screen_before = csi_fixed('1J')
    screen = csi_fixed('2J')
    line_after = csi_fixed('K')
    line_before = csi_fixed('1K')
    line = csi_fixed('2K')


class scroll(object):
    """
    ansi codes for scrolling
    """
    up = csi('S')
    down = csi('T')


def newline():
    write(csi('E'))


def writexy(xy, *args):
    """
    writes text on on screen
    a tuple as first argument gives the relative position to current TermCursor position
    does change TermCursor position
    args = list of optional position, formatting tokens and strings
    """
    write(TermCursor.moveby(*xy) + ''.join(args))


def put(xy, *args):
    """
    put text on on screen
    a tuple as first argument tells absolute position for the text
    does not change TermCursor position
    args = list of optional position, formatting tokens and strings
    """
    cmd = [TermCursor.save, TermCursor.move(*xy), ''.join(args), TermCursor.restore]
    write(''.join(cmd))


def getpassword(prompt="Password: "):
    """
    get user input without echo
    """

    fd = sys.stdin.fileno()
    old = termios.tcgetattr(fd)
    new = termios.tcgetattr(fd)
    new[3] &= ~termios.ECHO          # lflags
    try:
        termios.tcsetattr(fd, termios.TCSADRAIN, new)
        passwd = raw_input(prompt)
    finally:
        termios.tcsetattr(fd, termios.TCSADRAIN, old)
    return passwd


def getch():
    """
    get character. waiting for key
    """
    try:
        termios.tcsetattr(_fd, termios.TCSANOW, _new_settings)
        ch = sys.stdin.read(1)
    finally:
        termios.tcsetattr(_fd, termios.TCSADRAIN, _old_settings)
    return ch


# if __name__ == '__main__':
#     t = getpassword()
#     print t
#



# -*- coding: utf-8 -*-
# import all
from .progress import *
from .tictoc import *
from .dictobj import *
from .odict import *
from .flags import *
from .basic import *
try:
    from .jsonutil import *
except ImportError:
    pass
from .concurrent import *
from .tictoc import *
from .stringutil import *
from .iterflow import *

# from .listx import *

# import
from .progress import (progress, iprogress)
from .enum import Enum
from .named import (named, NamedMeta)
from .sizeof import sizeof

# modules
from .img import *
from .plt import *
from .io import *

# import like as module
from . import term
from . import decotool as deco
try:
    from . import database as db
except ImportError:
    pass
from . import irandom as rand

from .activeq import ActiveQ

# from . import queues as q



# -*- coding: utf-8 -*-
import matplotlib.pyplot as plt
from functools import wraps
from matplotlib._cm import datad

# reverse all the colormaps.
# reversed colormaps have '_r' appended to the name.

# http://matplotlib.org/examples/color/colormaps_reference.html
# _cmaps = [('Perceptually Uniform Sequential', ['viridis', 'plasma', 'inferno', 'magma']),
#          ('Sequential', ['Greys', 'Purples', 'Blues', 'Greens', 'Oranges', 'Reds',
#                          'YlOrBr', 'YlOrRd', 'OrRd', 'PuRd', 'RdPu', 'BuPu',
#                          'GnBu', 'PuBu', 'YlGnBu', 'PuBuGn', 'BuGn', 'YlGn']),
#          ('Sequential (2)', ['binary', 'gist_yarg', 'gist_gray', 'gray', 'bone', 'pink',
#                              'spring', 'summer', 'autumn', 'winter', 'cool', 'Wistia',
#                              'hot', 'afmhot', 'gist_heat', 'copper']),
#          ('Diverging', ['PiYG', 'PRGn', 'BrBG', 'PuOr', 'RdGy', 'RdBu',
#                         'RdYlBu', 'RdYlGn', 'Spectral', 'coolwarm', 'bwr', 'seismic']),
#          ('Qualitative', ['Pastel1', 'Pastel2', 'Paired', 'Accent', 'Dark2', 'Set1', 'Set2', 'Set3',
#                           'tab10', 'tab20', 'tab20b', 'tab20c']),
#          ('Miscellaneous', ['flag', 'prism', 'ocean', 'gist_earth', 'terrain', 'gist_stern',
#                             'gnuplot', 'gnuplot2', 'CMRmap', 'cubehelix', 'brg', 'hsv',
#                             'gist_rainbow', 'rainbow', 'jet', 'nipy_spectral', 'gist_ncar'])]


def cmap_fun(cm):

    def wrapper(x, alpha=None, bytes=False, lut=None, clim=None, norm=None):
        cmap = plt.get_cmap(cm, lut=lut)
        if norm is None:
            clim = clim or (0., 1.)
            norm = plt.Normalize(vmin=clim[0], vmax=clim[1])
        x = norm(x)

        return cmap(x, alpha=alpha, bytes=bytes)

    return wrapper


# explicte declaration (for example)
def jet(x, alpha=None, bytes=False, lut=None, clim=None, norm=None):
    return cmap_fun('jet')(x, alpha=alpha, bytes=bytes, lut=lut, clim=clim, norm=norm)


# fill functions
locals().update({cm: cmap_fun(cm) for cm in datad.keys()})
locals().update({cm+'_r': cmap_fun(cm + '_r') for cm in datad.keys()})





# -*- coding: utf-8 -*-
from contextlib import contextmanager
import numpy as np


@contextmanager
def np_seed(seed):
    """
    numpy random seed context
    :param seed:
    :return:
    """
    if seed is not None:
        state = np.random.get_state()
        np.random.seed(seed)
        yield
        np.random.set_state(state)

    else:
        yield


def rand_lazy_size(fun, **kw):

    def wrapper(*args, **kwargs):
        kwargs.update(kw)
        dtype = kwargs.pop('dtype', 'float32')

        def wrapped(shape):
            return fun(*args, size=shape).astype(dtype)
        return wrapped

    return wrapper

# most common only implemented
uniform = rand_lazy_size(np.random.uniform)
normal = rand_lazy_size(np.random.normal)
randint = rand_lazy_size(np.random.randint, dtype='int32')




# -*- coding: utf-8 -*-
try:
    import cjson as json
except Exception:
    import ujson as json


import numpy as np


def numpy_to_json(x):
    return json.encode(x.tolist())


def tojson(o):
    """
    recursive implementation
    """
    try:
        return json.encode(o)
    except json.EncodeError:
        pass
    try:
        return o.tojson()
    except AttributeError as e:
        pass

    t = type(o)
    if isinstance(o, list):
        return '[%s]' % ', '.join([tojson(e) for e in o])
    elif isinstance(o, dict):
        d = ['%s:%s' % (k, tojson(v)) for k, v in o.iteritems()]
        return '{%s}' % ', '.join(d)
    elif isinstance(o, set):
        d = ['%s:%s' % (tojson(e)) for e in o]
        return '{%s}' % ', '.join(d)
    elif isinstance(o, np.ndarray):
        return numpy_to_json(o)
    else:
        raise ValueError('error, failed encoding type(%s) to json' % t)




# -*- coding: utf-8 -*-
from . import img
from . import io
try:
    from . import plt
except NameError:
    # matplotlib not imported
    print('warning : matplotlib not imported')
    pass
from .flags import *




# -*- coding: utf-8 -*-
import functools
import inspect


def deco_optional(decorator):
    """
    optional argument 를 포함하는 decorator를 만드는 decorator
    """

    @functools.wraps(decorator)
    def dispatcher(*args, **kwargs):
        one_arg = len(args) == 1 and not kwargs
        if one_arg and inspect.isfunction(args[0]):
            decor_obj = decorator()
            return decor_obj(args[0])
        else:
            return decorator(*args, **kwargs)

    return dispatcher


def optional(deco):
    """
    decorator option은 kwargs만 허용
    :param deco:
    :return:
    """
    @functools.wraps(deco)
    def dispatcher(*args, **kwargs):
        decorator = deco(**kwargs)
        if args:
            assert len(args) == 1
            return decorator(args[0])
        else:
            return decorator
    return dispatcher


def optional_str(deco):
    """
    string 1개만 deco 인자로 올 수 있다.
    :param deco:
    :return:
    """

    @functools.wraps(deco)
    def dispatcher(*args, **kwargs):

        # when only function arg
        if not kwargs and len(args) == 1 \
                and not isinstance(args[0], str) \
                and args[0] is not None:
            decorator = deco()
            return decorator(args[0])
        else:
            # decorator with args
            decorator = deco(*args, **kwargs)
            return decorator

    return dispatcher


# noinspection PyProtectedMember
class bind(object):
    """
    # def bind(fun, *argsbind, **kwbind):
    #     if argsbind:
    #         return bindargs(fun, *argsbind, **kwbind)
    #     else:
    #         return bindkw(fun, **kwbind)
    #
    # # for bind function
    # bind._ = bind.placeholder = object()
    """
    _ = placeholder = object()

    def __new__(cls, fun, *argsbind, **kwbind):
        if argsbind:
            return bindargs(fun, *argsbind, **kwbind)
        else:
            return bindkw(fun, **kwbind)


def bindargs(fun, *argsbind, **kwbind):
    """
    _ = bind.placeholder   # unbound placeholder (arg)
    f = bind(fun, _, _, arg3, kw=kw1, kw2=kw2), f(arg1, arg2)
    :param fun:
    :param argsbind:
    :param kwbind:
    :return:
    """

    assert argsbind
    argsb = list(argsbind)
    iargs = [i for i in range(len(argsbind)) if argsbind[i] is bind.placeholder]
    # iargs = [a is bind.placeholder for a in argsbind]

    @functools.wraps(fun)
    def wrapped(*args, **kwargs):
        kws = kwbind.copy()

        args_this = [a for a in argsb]
        for i, arg in zip(iargs, args):
            args_this[i] = arg
        args_this.extend(args[len(iargs):])

        # kwargs.update(kwbind)
        kws.update(kwargs)
        # return fun(*argsb, **kws)
        return fun(*args_this, **kws)

    return wrapped


def bindkw(fun, **kwbind):
    """
    kwarg 바인딩된 함수 return.
    ex)
    def fun(opt1, opt2):
        print opt1, opt2

    f = bind(fun, opt1=2, opt2=3)
    f()
    :param function fun:
    :param kwbind:
    :return: function
    """

    @functools.wraps(fun)
    def wrapped(*args, **kwargs):
        kws = kwbind.copy()
        kws.update(kwargs)
        return fun(*args, **kws)

    return wrapped


class default_arg(object):
    """
    function decorator 디폴트 param control
    @default_arg(opt1=1, opt2=8)
    def example2(a, opt2, opt1):
        '''
        example2 function for test default decorator
        '''
        print a, opt1, opt2

    example2(1)  # print 1, 1, 8
    """

    def __init__(self, **kwargs):
        self.default = kwargs

    def __call__(self, fun):
        @functools.wraps(fun)
        def wrapped(*args, **kwargs):
            # default kwargs dict
            # wrapped.default = self.kwargs.copy()
            merge = wrapped.default.copy()
            merge.update(kwargs)

            return fun(*args, **merge)

        wrapped.default = self.default

        return wrapped


def default(fun, **kwdefault):
    """
    change default value for function
    ex)
    def sample(a, b=1, c=1):
        print 'from sample:', a, b, c
        return a, b, c
    fun = default(sample, b=4,c=5)
    print fun.default  # get default value dictionary
    fun(1)  # print 1, 5, 5 and return

    :param fun:
    :param kwdefault:
    :return:
    """

    @functools.wraps(fun)
    def wrapped(*args, **kwargs):
        merge = wrapped.default.copy()
        merge.update(kwargs)
        return fun(*args, **merge)

    wrapped.default = kwdefault

    return wrapped


def setup_once(initfn):
    """
    call class instance method for initial setup ::

        class B(object):

            def init(self, a):
                print 'init call:', a

            @setup_once(init)
            def mycall(self, a):
                print 'real call:', a

        b = B()
        b.mycall(222)
        b.mycall(333)

    :param function initfn:
    :return: decorated method
    """
    def wrap(method):

        finit = initfn.__name__
        fnname = method.__name__

        @functools.wraps(method)
        def wrapped(self, *args, **kwargs):

            @functools.wraps(method)
            def aftersetup(*a, **kw):
                return method(self, *a, **kw)

            setupfn = getattr(self, finit)
            setupfn(*args, **kwargs)

            res = method(self, *args, **kwargs)
            setattr(self, fnname, aftersetup)
            return res
        return wrapped

    return wrap


def call_once(method):

    fnname = method.__name__

    @functools.wraps(method)
    def wrapped(self, *args, **kwargs):

        @functools.wraps(method)
        def dummy(*_, **_kw):
            pass

        res = method(self, *args, **kwargs)

        setattr(self, fnname, dummy)
        return res
    return wrapped


def static(**kwargs):
    """ USE carefully ^^ """
    def wrap(fn):
        fn.func_globals['static'] = fn
        fn.__dict__.update(kwargs)
        return fn
    return wrap


@optional
def deprecated(message=''):
    import warnings

    def wrap(fun):
        @functools.wraps(fun)
        def wrapped(*args, **kwargs):
            warnings.warn("Call to deprecated function {}.".format(fun.__name__) + message,
                          DeprecationWarning)
            return fun(*args, **kwargs)
        return wrapped
    return wrap





# -*- coding: utf-8 -*-
"""thread and multiprocess common util functions
"""
from functools import wraps
from threading import Thread
from multiprocessing import Pool, Process
# from multiprocessing.queues import Queue
from multiprocessing import Queue


class SharedPool(object):
    """
    shared multiprocessing.pool instance
    """
    sharedpool = None

    @classmethod
    def get(cls, processes=None):
        if cls.sharedpool is None:
            cls.set(processes)
        return cls.sharedpool

    @classmethod
    def set(cls, processes=None):
        cls.sharedpool = Pool(processes)


# noinspection PyUnresolvedReferences
class ConcurrentCommon(object):
    """
    by multiprocess! not by thread
    res = Spawn(func, args...)
    print res.result  # wait for result..
    """

    def _maketarget(self, target):
        def _realtarget(*args, **kwargs):
            self.res.put(target(*args, **kwargs))

        return _realtarget

    @property
    def result(self):
        # make func interruptible
        while self.is_alive():
            self.join(0.01)
        return self.res.get()

    @property
    def done(self):
        return not self.is_alive()


class Threaded(Thread, ConcurrentCommon):
    """
    threaded call

    res = Threaded(func, args..., kwargs)
    print res.result  # wait for result..

    """

    def __init__(self, target, *args, **kwargs):
        target = self._maketarget(target)
        super(Threaded, self).__init__(target=target, args=args, kwargs=kwargs)
        self.daemon = True
        self.res = Queue(maxsize=1)
        self.start()


class Spawn(Process, ConcurrentCommon):
    """
    multiprocess call
    """

    def __init__(self, target, *args, **kwargs):
        target = self._maketarget(target)
        super(Spawn, self).__init__(target=target, args=args, kwargs=kwargs)
        self.daemon = True
        self.res = Queue(maxsize=1)
        self.start()


def threaded(f, *args, **kwargs):
    """function decorator
    """
    if args or kwargs:
        return Threaded(f, *args, **kwargs)

    @wraps(f)
    def wrapped(*wargs, **wkwargs):
        return Threaded(f, *wargs, **wkwargs)

    return wrapped


def spawn(f, *args, **kwargs):
    """decorator
    """
    if args or kwargs:
        return Spawn(f, *args, **kwargs)

    @wraps(f)
    def wrapped(*args, **kwargs):
        return Spawn(f, *args, **kwargs)

    return wrapped





# -*- coding: utf-8 -*-
import logging
import functools
import sys
from .decotool import optional
from .ansi import ansi
import collections


class FormatterX(logging.Formatter):

    """
    formatter extension.
    1) {} style formatting
    2) different format by level
    """

    def __init__(self, fmt=None, datefmt=None, **fmt_level):
        self.fmt_ = fmt or '{message}'
        self.fmt = fmt_level or dict()  # format by level
        self.datefmt = datefmt or "%m-%d %H:%M"

    def usesTime(self):
        return self._fmt.find("{asctime}") >= 0

    def getfmt(self, levelname):
        return self.fmt.get(levelname, self.fmt_)

    def setFormat(self, **kwargs):
        self.fmt.update(kwargs)

    def format(self, record):
        """tweaked from source of base"""
        try:
            record.message = record.getMessage()
        except TypeError:
            # if error during msg = msg % self.args
            if record.args:
                if isinstance(record.args, collections.Mapping):
                    record.message = record.msg.format(**record.args)
                else:
                    record.message = record.msg.format(record.args)
        self._fmt = self.getfmt(record.levelname)
        if self.usesTime():
            record.asctime = self.formatTime(record, self.datefmt)

        s = self._fmt.format(**record.__dict__)

        if record.exc_info:
            # Cache the traceback text to avoid converting it multiple times
            # (it's constant anyway)
            if not record.exc_text:
                record.exc_text = self.formatException(record.exc_info)
        if record.exc_text:
            if s[-1:] != '\n':
                s += '\n'
            try:
                s = s + record.exc_text
            except UnicodeError:
                s = s + record.exc_text.decode(sys.getfilesystemencoding(), 'replace')
        return s

# noinspection PyStatementEffect
default_format = {
    'DEBUG': ansi.yellow('[D] {asctime}: [{funcName}()] {message}'),
    'WARNING': ansi.light_yellow('[W] {asctime}: [{funcName}()] {message}'),
    'INFO': ansi.light_green('[I] {asctime}: [{funcName}()] {message}'),
    'ERROR': ansi.light_red('[E] {asctime}: [{filename}][{lineno}] [{funcName}()] {message}'),
    'CRITICAL': ansi.bold('[F] {asctime}: [{filename}][{lineno}] [{funcName}()] {message}') / ansi.dark_red
}

# common formatter, handler
formatter = FormatterX(**default_format)
default_handler = logging.StreamHandler()
default_handler.setFormatter(formatter)


class Loggerx(logging.Logger):

    def _log(self, *args, **kwargs):
        if self.disabled:
            return
        super(Loggerx, self)._log(*args, **kwargs)


# register logger class
logging.setLoggerClass(Loggerx)


def getlogger(pkg='', handler=None):
    """
    패키지 혹은 채널 로거
    logging.getLogger(package_name) or logg.getLogger()
    :param pkg: str
    """
    from .caller import caller

    if not pkg:
        m = caller.modulename()
        s = m.split('.', 1)
        if len(s) > 1:
            pkg = s[0]

    if haslogger(pkg):
        return logging.getLogger(pkg)
    else:
        # local
        logger = logging.getLogger(pkg)
        logger.addHandler(handler or default_handler)
        logger.setLevel(logging.DEBUG)
        return logger


def haslogger(name):
    """
    name의 로거가 있는지 없는지 체크. root logger는 무조건 만들어진다.
    """
    return name in logging.Logger.manager.loggerDict or not name


# function shortcuts
logg = getlogger('snipy')
info = logg.info
warn = logg.warn
debug = logg.debug
fatal = logg.critical
error = logg.error
exception = logg.exception


@optional
class trace(object):
    """decorator. 해당 함수 콜할때 마다 로그를 남기개 함.
    @trace
    def fx(x):
    등과 같이 쓰던지

    @trace(level=logging.INFO)
    def fx(x):
    등과 같이 사용하면 됨

    @trace(level=logging.INFO, logger='package')
    def fx(x):

    @trace(level=logging.INFO, logger=logger)
    def fx(x):

    데코레이터를 포함시킨 함수가 있는 모듈에 해당 모듈을 이름으로 하는 ogger 객체가 있다면.
    logger.log를, 없다면 logging.log를 호출함.
    """

    def __init__(self, level=logging.DEBUG, logger=None):
        self.level = level
        if logger is None or isinstance(logger, str):
            logger = getlogger(logger)
        self.logger = logger

    def __call__(self, func):
        from .caller import caller

        def make_message(args, kwargs):
            params = map(str, args) + ['%s=%s' % (k, v) for k, v in kwargs.items()]
            msg = "%s(%s) @ %s" % (func.__name__, ', '.join(params), caller.funname(depth=2))
            return msg

        @functools.wraps(func)
        def wrapped(*args, **kwargs):
            msg = make_message(args, kwargs)
            self.logger.log(self.level, msg)

            return func(*args, **kwargs)

        return wrapped


def basicConfig(**kw):
    """logging의 로그를 한번 호출하면 basicConfig가 안먹으므로. 기존 핸들러 삭제후 재설정.
    http://stackoverflow.com/questions/1943747/python-logging-before-you-run-logging-basicconfig
    ex)
    basicConfig(format='%(asctime)s %(message)s', level=logging.DEBUG)
    :param filename: Specifies that a FileHandler be created, using the specified filename, rather than a StreamHandler.
    :param filemode: Specifies the mode to open the file, if filename is specified (if filemode is unspecified, it defaults to ‘a’).
    :param format: Use the specified format string for the handler. (https://docs.python.org/2.7/library/logging.html#logging.basicConfig
    :param datefmt: Use the specified date/time format.
    :param level: Set the root logger level to the specified level.
    :param stream: Use the specified stream to initialize the StreamHandler. Note that this argument is incompatible with ‘filename’ - if both are present, ‘stream’ is ignored.

    """
    while len(logging.root.handlers) > 0:
        logging.root.removeHandler(logging.root.handlers[-1])

    logging.basicConfig(**kw)



# -*- coding: utf-8 -*-
from collections import (MutableMapping)

import argparse as _argparse
import snipy.ilogging as logg

# modified from tensorflow.python.platform.flags


def _str2bool(v):
    return v.lower() in ('yes', 'true', 't', '1')


def _str2eval(v):
    # list or tuple or str ...
    return eval(v)


# add other type parsing methods
_parser = _argparse.ArgumentParser()
_parser.register('type', 'bool', _str2bool)
_parser.register('type', 'list', _str2eval)
# _parser.register('type', 'tuple', _str2eval)


# region flag class

class _FlagValuesNone(MutableMapping):
    """Global container and accessor for flags and their values."""

    def __init__(self):
        self.__dict__['__flags'] = {}
        self.__dict__['__parsed'] = False

# region dictionaly interface

    def __getitem__(self, item):
        return self.__getattr__(item)

    def __setitem__(self, key, value):
        self.__dict__['__flags'][key] = value

    def __delitem__(self, key):
        del self.__dict__['__flags'][key]

    def __iter__(self):
        return iter(self.__dict__['__flags'])

    def __len__(self):
        return len(self.__dict__['__flags'])

    def __str__(self):
        return str(self.__dict__['__flags'])

    def __repr__(self):
        return self.__dict__['__flags'].__repr__()
# endregion

    def _parse_flags(self, args=None, verbose=True):
        silent = not verbose
        result, unparsed = _parser.parse_known_args(args=args)
        for flag_name, val in vars(result).items():
            self.__dict__['__flags'][flag_name] = val
            silent or logg.info('--{}={}'.format(flag_name, val))

        self.__dict__['__parsed'] = True
        # if unparsed:
        #     logg.info('unparsed : {}'.format(unparsed))
        return unparsed

    def _parse_flags_kw(self, args=None, verbose=True):
        left = self._parse_flags(args=args, verbose=verbose)
        unparsed = []
        kw = dict()
        while left:
            p = left.pop(0)
            if p.startswith('--'):
                s = p[2:].split('=')
                if len(s) == 2:
                    kw[s[0]] = s[1]
                else:
                    kw[s[0]] = left.pop(0)
            else:
                unparsed.append(p)

        # flag_not_parsed = tuple(f for f in unparsed if f.startswith('--'))
        if verbose and (unparsed or kw):
            if unparsed:
                logg.warn('as args: {}'.format(unparsed))
            if kw:
                logg.warn('as kwargs: {}'.format(kw))
        return unparsed, kw

    def __getattr__(self, name):
        """Retrieves the 'value' attribute of the flag --name."""
        if not self.__dict__['__parsed']:
            self._parse_flags()
        if name not in self.__dict__['__flags']:
            return None
        return self.__dict__['__flags'][name]

    def __setattr__(self, name, value):
        """Sets the 'value' attribute of the flag --name."""
        if not self.__dict__['__parsed']:
            self._parse_flags()
        self.__dict__['__flags'][name] = value

    def add_flag(self, flagname, default_value, help='', dtype=None, **kwargs):
        def gettype_default(v):
            if v is None:
                return str, v
            elif isinstance(v, bool):
                return 'bool', v
            elif type(v) is type:
                # 디폴트 대신 type이 들어옴. 디폴트 없는 것으로
                return v, None
            # elif isinstance(v, (tuple, list)):
            #     return
            else:
                return type(v), v

        h = help or flagname.replace('_', ' ')
        if dtype is None:
            dtype, default_value = gettype_default(default_value)

        h += ' : default = {}'.format(default_value)

        _parser.add_argument('--' + flagname, type=dtype,
                             default=default_value, help=h, **kwargs)

        self.__dict__['__parsed'] = False

    def add_args(self, args):
        # add positional arguments
        for i, a in enumerate(args):
            _parser.add_argument(a, nargs=i+1)

flag = _FlagValuesNone()

# endregion


# def remove_flag(name):
#     """remove flag from parser"""
#     for action in _parser._actions:
#         if vars(action)['option_strings'][0] == name:
#             _parser._handle_conflict_resolve(None, [(name, action)])
#             break


def add_flag(*args, **kwargs):
    """
    define a single flag.
    add_flag(flagname, default_value, help='', **kwargs)
    add_flag([(flagname, default_value, help), ...])
    or
    define flags without help message
    add_flag(flagname, default_value, help='', **kwargs)

    add_flag('gpu', 1, help='CUDA_VISIBLE_DEVICES')
    :param args:
    :param kwargs:
    :return:
    """
    if len(args) == 1 and isinstance(args[0], (list, tuple)):
        for a in args[0]:
            flag.add_flag(*a)
    elif args:
        flag.add_flag(*args, **kwargs)
    else:
        for f, v in kwargs.items():
            flag.add_flag(f, v)


def parse_flag(*args, **kwargs):
    return flag._parse_flags(*args, **kwargs)



# region run

def run(main=None, argv=None, **flags):
    """
    :param main: main or sys.modules['__main__'].main
    :param argv: argument list used in argument parse
    :param flags: flags to define with defaults
    :return:
    """
    """Runs the program with an optional 'main' function and 'argv' list."""
    import sys as _sys
    import inspect
    main = main or _sys.modules['__main__'].main

    if main.__doc__:
        docstring = main.__doc__.split(':param')[0]
        _parser.usage = 'from docstring \n {}'.format(docstring)  # add_help

    # if not flags:
    try:
        a = inspect.getfullargspec(main)
    except AttributeError:
        a = inspect.getargspec(main)  # namedtuple(args, varargs, keywords, defaults)
    if a.defaults:
        kwargs = dict(zip(reversed(a.args), reversed(a.defaults)))
        add_flag(**kwargs)
    else:
        kwargs = dict()

    # add to command argument
    if a.defaults is None:
        nargs = len(a.args)
    else:
        nargs = len(a.args) - len(a.defaults)
    # if nargs > 0:
    posargs = a.args[:nargs]
    flag.add_args(posargs)
    add_flag(**flags)

    # Extract the args from the optional `argv` list.
    args = argv[1:] if argv else None

    # Parse the known flags from that list, or from the command
    # line otherwise.
    unparsed, kw = flag._parse_flags_kw(args=args)

    d = flag.__dict__['__flags']
    args = [d[k] for k in posargs]
    args += unparsed

    kwargs.update({k: d[k] for k in kwargs.keys()})
    kwargs.update(kw)

    # Call the main function, passing through any arguments,
    # with parsed flags as kwwargs
    # to the final program.
    _sys.exit(main(*args, **kwargs))

# endregion



# -*- coding: utf-8 -*-
from collections import Mapping


class DictObj(dict):
    """
    object attribute 혹은 dictionary처럼 접근하는 클래스
    """

    def __init__(self, *args, **kwargs):
        super(DictObj, self).__init__(*args, **kwargs)
        self.__dict__ = self

    def __getattr__(self, item):
        """
        해당 키가 없을 때 None을 준다.
        :param item: 키
        :rtype None:
        """
        return None

    def __sub__(self, other):
        """
        :type other: dict
        :rtype: dictobj: other dict에 없는 items만 리턴
        """
        return DictObj({k: v for k, v in self.items() if k not in other})

    def __add__(self, other):
        if not other:
            return self.copy()
        added = self.copy()
        added.update(other)
        return added

    __radd__ = __add__

    def __getnewargs__(self):
        """pickling related"""
        return tuple()

    def __getstate__(self):
        return self.__dict__

    def __setstate__(self, d):
        self.__dict__.update(d)

    def copy(self):
        return DictObj(super(DictObj, self).copy())

    def intersect(self, other):
        """
        self와 other 키가 동일한 아이템의 dictobj
        :type other: dict
        :rtype: dictobj:
        """
        return DictObj({k: self[k] for k in self if k in other})

    @staticmethod
    def convert_ifdic(value):
        if isinstance(value, Mapping):
            return DictObj.from_dict(value)
        else:
            return value

    @staticmethod
    def from_dict(dic):
        """
        recursive dict to dictobj 컨버트
        :param dic:
        :return:
        """
        return DictObj({k: DictObj.convert_ifdic(v) for k, v in dic.items()})


# alias
dictobj = DictObj



# -*- coding: utf-8 -*-
import time
from contextlib import contextmanager

from . ilogging import logg


def tic():
    """
    ex1)
    tic()  # save start time - time1
    toc()  # print elapsed time from last calling tic()
    toc()  # print elapsed time from last calling tic()

    ex2)
    t0 = tic()  # simple
    t1 = tic()
    toc(t1)  # print time from t1
    toc(t0)  # print time from t0

    :return: time:
    """
    t = time.time()
    setattr(tic, 'last_tic_time', t)
    return t


# noinspection PyUnresolvedReferences
def toc(t=None, name='tictoc'):
    """
    ex1)
    tic()  # save start time - time1
    toc()  # print elapsed time from last calling tic()
    toc()  # print elapsed time from last calling tic()

    ex2)
    t0 = tic()  # simple
    t1 = tic()
    toc(t1)  # print time from t1
    toc(t0)  # print time from t0

    :param t: time: 시작 시간 (tic()의 리턴 값)
    :param name: str: 출력시 포함할 문자 ['tictoc']
    """
    try:
        t = t or tic.last_tic_time
    except AttributeError:
        # tic()부터 콜하세요
        logg.warn('calling tic() need to use toc()')
        return
    elapsed = time.time() - t
    logg.info('%s Elapsed: %s secs' % (name, elapsed))
    return elapsed


@contextmanager
def tictoc(name='tictoc'):
    """
    with tictoc('any string or not'):
        print 'cool~~~'
    cool~~~
    2015-12-30 14:39:28,458 [INFO] tictoc Elapsed: 7.10487365723e-05 secs
    :param name: str
    """
    t = time.time()
    yield
    logg.info('%s Elapsed: %s secs' % (name, time.time() - t))




# -*- coding: utf-8 -*-
try:
    import Queue as PQ
except ImportError:  # ModuleNotFoundError:
    import queue as PQ

from random import randrange

import errno

from .concurrent import Threaded


class QueueRandom(PQ.Queue):

    """
    queue class for random put or get
    note: not a multiprocess.queue
    """

    def __init__(self, producer=None, maxsize=1024, processor=None):
        # old style class
        PQ.Queue.__init__(self, maxsize=maxsize)
        self.producer = producer
        self.proc = processor
        self.th = None
        self.produce()

    def _init(self, maxsize):
        self.queue = []

    def _produce(self, producer):
        if self.proc:
            for d in producer:
                self.put(self.proc(d))
        else:
            for d in producer:
                self.put(d)

    def produce(self):
        self.th = Threaded(self._produce, self.producer)

    def _get(self):
        # if empty?
        return self.queue.pop(randrange(0, self._qsize()))

    def _put(self, item):
        self.queue.insert(randrange(0, self._qsize() + 1), item)

    def __iter__(self):
        if self.done:
            self.reset()
        return self
        # self.produce()
        # while self.th.is_alive() and not self.full():
        #     pass
        # while not self.done:
        #     yield self.get()

    def next(self):
        while self.th.is_alive() and not self.full():
            pass
        if not self.done:
            return self.get(block=True)
        else:
            raise StopIteration

    # def reset(self):
    #     self.produce()
    reset = produce

    @property
    def done(self):
        if self.th is None:
            return True
        return self.empty() and not self.th.is_alive()


from time import sleep
import threading
import multiprocessing.queues as MQ


def is_main_alive():
    """
    is 메인 쓰레드 alive?
    :rtype: bool
    """

    for t in threading.enumerate():
        if t.name == 'MainThread':
            return t.is_alive()
    print('MainThread not found')
    return False


class QueueInterruptable(MQ.Queue):
    """
    thread 구현용 put에서 blocking 시키지 않음
    """
    Empty = MQ.Empty
    Full = MQ.Full

    def __init__(self, maxsize=0):
        self.finish = False
        super(QueueInterruptable, self).__init__(maxsize=maxsize)

    def get(self, block=True, timeout=None, verbose=True):
        if not block:
            return super(QueueInterruptable, self).get(block=False, timeout=timeout)

        # ignore blocking options for not blocking
        while not self.finish and is_main_alive():
            try:
                data = super(QueueInterruptable, self).get(block=False, timeout=0.01)
                return data
            except MQ.Empty as e:
                # raise e
                sleep(0.01)
        # if verbose:
        #     print('Interrupted or stopped. OK! finish in get')
        # if not is_main_alive():
        #     self.close()
        self.if_no_main_alive_close()

    def put(self, item, block=True, timeout=None, verbose=True):
        # ignore blocking options for not blocking
        while not self.finish and is_main_alive():
            try:
                super(QueueInterruptable, self).put(item, False, timeout=0.01)
                return True
            except MQ.Full:
                sleep(0.01)
        # if verbose:
        #     print('Interrupted or stopped. OK! finish in put')
        self.if_no_main_alive_close()
        return False

    def if_no_main_alive_close(self):
        try:
            if not is_main_alive():
                self.close()
        except IOError as e:
            if e.errno == errno.EPIPE:
                pass

    def stop(self):
        self.finish = True

    def close(self):
        super(QueueInterruptable, self).close()
        self.join_thread()




# -*- coding: utf-8 -*-

"""
https://en.wikipedia.org/wiki/ANSI_escape_code
http://wiki.bash-hackers.org/scripting/terminalcodes
http://ascii-table.com/ansi-escape-sequences.php
http://www.ecma-international.org/publications/files/ECMA-ST/Ecma-048.pdf
"""


def ansi_str(s, fmt):
    if isinstance(s, str):
        return '\x1b[%sm%s\x1b[0m' % (fmt, s)
    else:
        return '\x1b[%sm%s\x1b[0m' % (fmt, str(s))


def ansi_str_restore(s, fmt, patt='\x1b[0m'):
    fmt2 = '\x1b[0;%sm' % fmt
    a = str(s).replace(patt, fmt2)

    return '\x1b[%sm%s\x1b[0m' % (fmt, a)


class _AnsiBuilder(object):
    def __init__(self, fmt):
        self.fmt = fmt

    def __call__(self, s):
        """ make ansi string
        """
        return ansi_str_restore(s, self.fmt)

    __radd__ = __call__  # same

    def __add__(self, other):
        """ make ansi string (simple ver.)
        """
        return ansi_str(other, self.fmt)

    __rmul__ = __rdiv__ = __add__

    def __mul__(self, other):
        """ merge ansi format
        """
        self.fmt += ';' + other.fmt
        return self


class _AttrAnsi(_AnsiBuilder):
    def __mul__(self, other):
        return _AnsiBuilder(self.fmt) * other


class _ColorAnsi(object):
    def __init__(self, c, fmt=''):
        self.c = c
        self.fg = fmt + str(30 + c)
        self.bg = fmt + str(40 + c)

    def __call__(self, s):
        return ansi_str_restore(s, self.fg)

    __radd__ = __call__

    def __add__(self, other):
        return ansi_str(other, self.fg)

    def __mul__(self, other):
        return _AnsiBuilder(self.fg) * other

    def __rmul__(self, other):
        return other * _AnsiBuilder(self.fg)

    def __div__(self, other):
        return _AnsiBuilder(self.fg) / other

    def __rtruediv__(self, other):
        # python3
        return _AnsiBuilder(self.fg) / other

    def __rdiv__(self, other):
        return other * _AnsiBuilder(self.bg)

    def __rtruediv__(self, other):
        return other * _AnsiBuilder(self.bg)

    @property
    def fmt(self):
        return self.fg


class _NoAnsi(object):
    def __init__(self):
        pass

    def __call__(self, s):
        # if isinstance(s, (str, basestring)):  # python2 only
        if isinstance(s, str):
            return s
        else:
            return str(s)

    __radd__ = __add__ = __rmul__ = __rdiv__ = __call__

    def __mul__(self, _):
        return self

    __div__ = __mul__


noAnsi = _NoAnsi()


class ansi(object):

    attrName = ['normal', 'bold', 'dark', 'italic', 'underline', 'blink', 'rapid', 'negative']
    colorName = ['black', 'red', 'green', 'yellow', 'blue', 'magenta', 'cyan', 'white']

    normal = _AttrAnsi('0')
    bold = _AttrAnsi('1')
    dark = _AttrAnsi('2')

    italic = _AttrAnsi('3')
    underline= _AttrAnsi('4')
    blink = _AttrAnsi('5')
    rapid = _AttrAnsi('6')
    negative = _AttrAnsi('7')
    light = bold
    #blink = attr('5')

    black = _ColorAnsi(c=0)
    red = _ColorAnsi(c=1)
    green = _ColorAnsi(c=2)
    yellow = _ColorAnsi(c=3)
    blue = _ColorAnsi(c=4)
    magenta = _ColorAnsi(c=5)
    cyan = _ColorAnsi(c=6)
    white = _ColorAnsi(c=7)

    light_black = _ColorAnsi(c=0, fmt='1;')
    light_red = _ColorAnsi(c=1, fmt='1;')
    light_green = _ColorAnsi(c=2, fmt='1;')
    light_yellow = _ColorAnsi(c=3, fmt='1;')
    light_blue = _ColorAnsi(c=4, fmt='1;')
    light_magenta = _ColorAnsi(c=5, fmt='1;')
    light_cyan = _ColorAnsi(c=6, fmt='1;')
    light_white = _ColorAnsi(c=7, fmt='1;')

    dark_black = _ColorAnsi(c=0, fmt='2;')
    dark_red = _ColorAnsi(c=1, fmt='2;')
    dark_green = _ColorAnsi(c=2, fmt='2;')
    dark_yellow = _ColorAnsi(c=3, fmt='2;')
    dark_blue = _ColorAnsi(c=4, fmt='2;')
    dark_magenta = _ColorAnsi(c=5, fmt='2;')
    dark_cyan = _ColorAnsi(c=6, fmt='2;')
    dark_white = _ColorAnsi(c=7, fmt='2;')

    @staticmethod
    def colors():
        return [ansi.__dict__[c] for c in ansi.colorName]

    @staticmethod
    def ansiattr():
        return [ansi.__dict__[c] for c in ansi.attrName]


def enable_color():
    d = dict()

    for i, name in enumerate(ansi.attrName):
        d[name] = _AttrAnsi(str(i))

    d['light'] = d['bold']

    for i, name in enumerate(ansi.colorName):
        d[name] = _ColorAnsi(c=i)
        d['light_' + name] = _ColorAnsi(c=i, fmt='1;')
        d['dark_' + name] = _ColorAnsi(c=i, fmt='2;')

    for k, v in d.items():
        setattr(ansi, k, v)
    globals().update(d)


def disable_color():
    d = dict()

    for i, name in enumerate(ansi.attrName):
        d[name] = noAnsi

    d['light'] = noAnsi

    for i, name in enumerate(ansi.colorName):
        d[name] = noAnsi
        d['light_' + name] = noAnsi
        d['dark_' + name] = noAnsi

    for k, v in d.items():
        setattr(ansi, k, v)
    globals().update(d)


global normal, bold, dark, italic, underline, blink, rapid, negative, light
global black, red, green, yellow, blue, magenta, cyan, white
global light_black, light_red, light_green, light_yellow, light_blue, light_magenta, light_cyan, light_white
global dark_black, dark_red, dark_green, dark_yellow, dark_blue, dark_magenta, dark_cyan, dark_white




# -*- coding: utf-8 -*-
import collections


def named(typename, *fieldnames, **defaults):
    """
    namedtuple with default values
    named('typename', fields | *fields, default=x, [**defaults])
    :param typename:
    :param fieldnames:
    :param defaults:
    :return:
    """
    if len(fieldnames) == 1:
        if isinstance(fieldnames[0], str):
            fieldnames = tuple(fieldnames[0].replace(',', ' ').split())
        elif isinstance(fieldnames[0], (list, tuple)):
            fieldnames = fieldnames[0]

    # set default of defaults
    default_of_defaults = defaults.pop('default', None)

    dfields = tuple(f for f in defaults if f not in fieldnames)

    T = collections.namedtuple(typename, fieldnames + dfields)
    T.__new__.__defaults__ = (default_of_defaults,) * len(T._fields)
    prototype = T(**defaults)
    T.__new__.__defaults__ = tuple(prototype)

    # make picklable
    globals()[typename] = T

    return T


class NamedMeta(type):

    def __new__(mcs, name, bases, dct):
        """ contains default values,
        dct contains __slots__ for being ordered
        bases is ignored
        """

        slots = dct.get('__slots__', ())
        defaults = dct.get('__defaults__', {})

        # defaults in base class arguments ^^
        # otherwise ignoring
        for base in bases:
            if isinstance(base, collections.Mapping):
                defaults.update(base)
            else:
                continue

        # if need inheritance?,
        # IDE cannot resolve attributes.
        # below is experiment for inheritance.
        #
        # for base in bases:
        #     # get slots of base
        #     slot = base._fields
        #     default = base.__new__.__defaults__
        #     defaults.update({s: v for s, v in zip(slot, default)})
        #     slots += slot

        return named(name, *slots, **defaults)


# region sample for namedmeta class

# class SampleNamed:
#     __metaclass__ = NamedMeta
#     __slots__ = ('prop1', 'prop2')
#     __defaults__ = dict(prop1=1, prop2=2, default=1)
#
#
# def test_named_default():
#     a = SampleNamed()
#     print a.prop1
#     print a.prop2
#     print a
#
#     import joblib
#     joblib.dump(a, '/tmp/named_default.pkl')
#     a2 = joblib.load('/tmp/named_default.pkl')
#     print a2
#     print isinstance(a, (tuple))

# endregion


# region experimental

# class NamedTestMeta(type):
#
#     def __new__(mcs, name, bases, dct):
#         """bases contains default values,
#         dct contains __slots__ for being ordered
#         """
#         # assert len(bases) == 1 and isinstance(bases[0], collections.Mapping)
#         slots = dct.get('__slots__', ())
#         defaults = {}
#         for base in bases:
#             if isinstance(base, collections.Mapping):
#                 defaults.update(base)
#             elif isinstance(base, (tuple, list)):
#                 slots += tuple(base)
#             elif isinstance(base, type):
#                 continue
#             else:
#                 slots += (base,)
#
#         return named(name, *slots, **defaults)
#
#
# class SampleTuplet('a b c', dict(a=2, b=3)):
#     __metaclass__ = NamedTestMeta
#     # __slots__ = ('a', 'b', 'c')

# endregion





# -*- coding: utf-8 -*-
import numpy as np
from ..irandom import np_seed


def split_rand(data_or_size, ratio, seed):
    """
    data(1-ratio), data(with ratio) = split_rand(data_or_size, ratio, seed)
    :param data_or_size: data or count
    :param ratio:
    :param seed:
    :return:
    """
    if not isinstance(data_or_size, int):
        sz = len(data_or_size)
        data = np.asarray(data_or_size)
    else:
        sz = data_or_size
        data = np.arange(sz)
    if not ratio:
        return data, []

    i = np.zeros(sz, dtype='bool')
    lattersz = int(sz * ratio)
    i[:lattersz] = True

    with np_seed(seed):
        np.random.shuffle(i)

    return data[~i], data[i]


def testsets(data_or_sz, p_testset=None, seed=7238):
    """

    :param data_or_sz:
    :param p_testset:
    :param seed:
    :return:
    """
    trains, tests = split_rand(data_or_sz, p_testset, seed)
    return tests, trains


def kfolds(n, k, sz, p_testset=None, seed=7238):
    """
    return train, valid  [,test]
    testset if p_testset
    :param n:
    :param k:
    :param sz:
    :param p_testset:
    :param seed:
    :return:
    """
    trains, tests = split_rand(sz, p_testset, seed)

    ntrain = len(trains)

    # np.random.seed(seed)
    with np_seed(seed):
        np.random.shuffle(trains)
    if n == k:
        # no split
        train, valid = trains, trains
    else:
        foldsz = ntrain // k

        itrain = np.arange(ntrain) // foldsz != n
        train = trains[itrain]
        valid = trains[~itrain]

    if not p_testset:
        return train, valid
    else:
        return train, valid, tests


if __name__ == '__main__':
    out = kfolds(2, 4, 100, 0.2)
    print out
    print map(len, out)




# -*- coding: utf-8 -*-


class EnumMeta(type):

    def __iter__(cls):
        return (k for k in cls.__dict__ if not k.startswith('__'))

    def iteritems(cls):
        return ((k, v) for k, v in cls.__dict__.items() if not k.startswith('__'))

    def values(cls):
        return [v for k, v in cls.iteritems()]

    def __contains__(cls, item):
        return (item in iter(cls)) or (item in cls.values())

    def __getitem__(self, item):
        return getattr(self, item)


class Enum(object):
    __metaclass__ = EnumMeta


def enum(name, *members, **withvalue):
    """class buider"""
    if len(members) == 1:
        if isinstance(members[0], str):
            members = members[0].split()
        elif isinstance(members[0], (list, tuple)):
            members = members[0]

    dic = {v: v for v in members}
    dic.update(withvalue)

    return type(name, (Enum,), dic)


if __name__ == '__main__':
    # Fruits = enum('Fruits', ['apple', 'banana'])
    class Fruits(Enum):
        banana = 1
        apple = 2

    for f in Fruits:
        print(f)
    print(1 in Fruits)




# -*- coding: utf-8 -*-
"""
reference : https://github.com/pypa/sampleproject/blob/master/setup.py
"""

from setuptools import (setup, find_packages)
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()


setup(name='snipy',
      version='0.0.7',
      description='python utility @ your own risk',
      long_description='snippet-like code to make snippet-like code',

      # The project's main homepage.
      url='https://github.com/dade-ai/snipy',
      author='dade',
      author_email='aiplore@gmail.com',

      license='MIT',

      # The project's main homepage.

      # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
      classifiers=[
          # How mature is this project? Common values are
          #   3 - Alpha
          #   4 - Beta
          #   5 - Production/Stable
          'Development Status :: 3 - Alpha',

          # Indicate who your project is intended for
          'Intended Audience :: Developers',

          # Pick your license as you wish (should match "license" above)
          'License :: OSI Approved :: MIT License',

          # Specify the Python versions you support here. In particular, ensure
          # that you indicate whether you support Python 2, Python 3 or both.
          'Programming Language :: Python :: 3.6',
      ],

      packages=find_packages(),
      install_requires=[
            'numpy',
            'setuptools',
      ],
      extras_require={
            'all': ['Jinja2'],
            'plot': ['scikit-image', 'matplotlib'],
            'image': ['scikit-image', 'Pillow', 'scikit-image'],
            'file': ['joblib', 'scandir'],
            'perlin': ['noise'],
            'h5': ['h5py'],
            'json': ['ujson'],
            'sizeof': ['pympler'],
            'cuda': ['pycuda']
      }
      )



# -*- coding: utf-8 -*-
import snipy.common as sp


def test_odict():
    data = [('a', 0), ('b', 1)]
    d = sp.odict(data)

    assert d['a'] == d.a == data[0][1]
    assert d['b'] == d.b == data[1][1]

    d.c = 2121
    assert d['c'] == d.c == 2121
    print(list(d.keys()))
    assert list(d.keys()) == 'a b c'.split()


if __name__ == '__main__':
    test_odict()




# -*- coding: utf-8 -*-


#region List class blocks


class ListObj(list):
    pass


class ListArgs(list):
    """
    list extention, can initialize by *args
    """
    __slots__ = ()

    def __init__(self, *args):
        try:
            # args가 1개이하이고, iterable 이라면
            super(ListArgs, self).__init__(*args)
        except TypeError:
            # tuple(args)들로 초기화
            super(ListArgs, self).__init__(args)


# noinspection PyCallingNonCallable
class ListCallable(ListArgs):
    """
    list content as callable-chain
    """
    __slots__ = ()

    def __call__(self, *args, **kwargs):
        res = self[0](*args, **kwargs)
        for call in self[1:]:
            res = call(res)
        return res


class ListAddable(ListArgs):
    """
    list addable (add(+) = extend or append)
    """
    __slots__ = ()

    def add(self, list_or_not):
        # self.add(other) = self[1,2,3] + other[4,5,6] => [1,2,3, 4,5,6]
        try:
            self.extend(list_or_not)
        except TypeError:
            self.append(list_or_not)
        return self

    def radd(self, other):
        # self.radd(other) = other[4,5,6] + self[1,2,3]  => [4,5,6, 1,2,3]
        try:
            self[:0] = other
        except TypeError:
            self.insert(0, other)
        return self

    def __add__(self, other):
        return self.__class__(self).add(other)

    def __radd__(self, other):
        return self.__class__(self).radd(other)

    def __iadd__(self, other):
        return self.add(other)


class ListCallChain(ListAddable, ListCallable):
    """
    list.add(extend or append) by >> operator
    """

    # self[1,2,3] >> other[4,5,6] => [1,2,3, 4,5,6]
    # other >> self => self(other)
    # self[1,2,3] << other[4,5,6] => [4,5,6, 1,2,3]
    # other << self => self.add(other)
    __slots__ = ()

    def __rshift__(self, other):
        # self[1,2,3] >> other[4,5,6] => [1,2,3, 4,5,6]
        # self >> other : lazy calling sequence add
        return self.__add__(other)

    def __rrshift__(self, other):
        # other >> self => self(other)
        return self.__call__(other)

    # << support
    def __lshift__(self, other):
        # self << other => as call : self(other)
        if hasattr(other, '__rlshift__'):
            return other.__rlshift__(self)
        else:
            return self.__call__(other)

    # << support
    def __rlshift__(self, other):
        # other << self => append other [self, other]
        return self.__add__(other)


class ICall(object):

    __slots__ = ()

    def __rrshift__(self, other):
        # other >> self => self(other)
        return self.__call__(other)

    # << support
    def __lshift__(self, other):
        # self << other => as call : self(other)
        if hasattr(other, '__rlshift__'):
            return other.__rlshift__(self)
        else:
            return self.__call__(other)


class IChain(object):

    __slots__ = ()

    def __rshift__(self, other):
        # self >> other
        # self.add(other) 보다 나은점은.
        # ichain >> ichainlist 내가 리스트일 필요는 없지만,
        # chain list class type을 모른다. other 쪽에서 정하게 하자.
        # other 가 _radd 구현해야할 책임이
        # checkme if same type??

        return other.__radd__(self)

    # << support not in this, other decide what to do
    def __rlshift__(self, other):
        # other << self
        return other.__radd__(self)


class ICallChain(ICall, IChain):

    __slots__ = ()

    def __radd__(self, other):
        """ overidden in LayerBasic
        other >> self
        """
        return ListCallChain(other, self)


# aliases
listobj = ListObj
listargs = ListArgs
listcallable = ListCallable
listaddable = ListAddable
listchain = ListCallChain


# def ListX(bases, clsname=''):
#     """list class factory"""
#     clsname = clsname or '_'.join(b.__name__ for b in bases)
#     return type(clsname, bases, {})

#endregion


if __name__ == '__main__':
    f = [lambda x:x+1, lambda x:x+2, lambda y: y+3]
    l = listchain(f)
    ll = l >> (lambda x: x + 4)
    print(0 >> l)
    print(l(0))
    print(ll(0))


